"""
Module to provide processing for the block quotes.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import (
    BlockQuoteStackToken,
    HtmlBlockStackToken,
    IndentedCodeBlockStackToken,
    LinkDefinitionStackToken,
    ListStackToken,
    ParagraphStackToken,
    TableBlockStackToken,
)

# pylint: disable=too-many-lines


POGGER = ParserLogger(logging.getLogger(__name__))


class BlockQuoteCountHelper:
    """
    Class to provide processing for the block quotes.
    """

    __block_quote_character = ">"

    @staticmethod
    def is_block_quote_start(
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        adj_ws: Optional[str] = None,
    ) -> bool:
        """
        Determine if we have the start of a block quote section.
        """

        return TabHelper.is_length_less_than_or_equal_to(
            extracted_whitespace if adj_ws is None else adj_ws, 3
        ) and ParserHelper.is_character_at_index(
            line_to_parse, start_index, BlockQuoteCountHelper.__block_quote_character
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def count_block_quote_starts(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        block_quote_data: BlockQuoteData,
        is_top_of_stack_fenced_code_block: bool,
        is_top_of_stack_is_html_block: bool,
    ) -> Tuple[BlockQuoteData, int, str, int, bool]:
        """
        Having detected a block quote character (">") on a line, continue to consume
        and count while the block quote pattern is there.
        """

        (
            last_block_quote_index,
            avoid_block_starts,
            adjusted_line,
        ) = (
            -1,
            False,
            line_to_parse,
        )
        if block_quote_data.stack_count == 0 and is_top_of_stack_fenced_code_block:
            start_index -= 1
        else:
            osi, oltp, current_count = start_index, line_to_parse[:], 1
            POGGER.debug("cbqs1-current_count--$--", current_count)
            start_index += 1

            POGGER.debug(
                "block_quote_data.stack_count--$--is_top_of_stack_fenced_code_block--$",
                block_quote_data.stack_count,
                is_top_of_stack_fenced_code_block,
            )

            last_block_quote_index = start_index
            while True:
                (
                    adjusted_line,
                    start_index,
                ) = BlockQuoteCountHelper.__handle_bq_whitespace(
                    adjusted_line, start_index
                )

                (
                    continue_processing,
                    avoid_block_starts,
                    start_index,
                    adjusted_line,
                    last_block_quote_index,
                    current_count,
                ) = BlockQuoteCountHelper.__should_continue_processing(
                    parser_state,
                    current_count,
                    block_quote_data.stack_count,
                    is_top_of_stack_is_html_block,
                    adjusted_line,
                    start_index,
                    osi,
                    oltp,
                    is_top_of_stack_fenced_code_block,
                    avoid_block_starts,
                    last_block_quote_index,
                )
                POGGER.debug("cbqs2-current_count--$--", current_count)
                if not continue_processing:
                    break
                current_count += 1
                POGGER.debug("cbqs3-current_count--$--", current_count)
                start_index += 1
                last_block_quote_index = start_index

            block_quote_data = BlockQuoteData(
                current_count, block_quote_data.stack_count
            )
            POGGER.debug(
                "__count_block_quote_starts--$--$--",
                start_index,
                adjusted_line,
            )
        return (
            block_quote_data,
            start_index,
            adjusted_line,
            last_block_quote_index,
            avoid_block_starts,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_bq_whitespace(adjusted_line: str, start_index: int) -> Tuple[str, int]:
        if ParserHelper.is_character_at_index_whitespace(adjusted_line, start_index):
            start_index += 1
        return adjusted_line, start_index

    # pylint: disable=too-many-arguments
    @staticmethod
    def __xx(
        parser_state: ParserState,
        adjusted_line: str,
        start_index: int,
        current_count: int,
        stack_count: int,
        last_block_quote_index: int,
    ) -> Tuple[bool, int, int, int]:
        (
            continue_processing,
            start_index,
        ) = BlockQuoteCountHelper.__is_special_double_block_case(
            parser_state, adjusted_line, start_index, current_count, stack_count
        )
        if not continue_processing and current_count < stack_count:
            continue_proc, stack_token_index = BlockQuoteCountHelper.__xx_part_one(
                parser_state, start_index, current_count
            )
            if continue_proc:
                current_count, start_index, last_block_quote_index = (
                    BlockQuoteCountHelper.__xx_part_two(
                        parser_state,
                        stack_token_index,
                        start_index,
                        current_count,
                        stack_count,
                        last_block_quote_index,
                    )
                )
        return continue_processing, current_count, start_index, last_block_quote_index

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __should_continue_processing(
        parser_state: ParserState,
        current_count: int,
        stack_count: int,
        is_top_of_stack_is_html_block: bool,
        adjusted_line: str,
        start_index: int,
        osi: int,
        oltp: str,
        is_top_of_stack_fenced_code_block: bool,
        avoid_block_starts: bool,
        last_block_quote_index: int,
    ) -> Tuple[bool, bool, int, str, int, int]:
        continue_processing = True
        POGGER.debug(
            "current_count--$--stack_count--$--is_top_of_stack_is_html_block--$",
            current_count,
            stack_count,
            is_top_of_stack_is_html_block,
        )
        if is_top_of_stack_is_html_block and current_count >= stack_count:
            if current_count == stack_count:
                POGGER.debug(
                    "block quote levels don't increase during html block, ignoring"
                )
                avoid_block_starts = ParserHelper.is_character_at_index(
                    adjusted_line,
                    start_index,
                    BlockQuoteCountHelper.__block_quote_character,
                )
                POGGER.debug("avoid_block_starts=$", avoid_block_starts)
                continue_processing = False
            else:
                assert current_count > stack_count, "Took care of ==, so left with >."
                (
                    start_index,
                    adjusted_line,
                    last_block_quote_index,
                    avoid_block_starts,
                    current_count,
                    continue_processing,
                ) = (
                    osi,
                    oltp,
                    -1,
                    True,
                    stack_count,
                    False,
                )

        if continue_processing:
            continue_processing = False
            if is_top_of_stack_fenced_code_block and (current_count >= stack_count):
                POGGER.debug("out of stack")
            elif start_index == len(adjusted_line):
                POGGER.debug("ran out of line")
            elif ParserHelper.is_character_at_index_not(
                adjusted_line,
                start_index,
                BlockQuoteCountHelper.__block_quote_character,
            ):
                (
                    continue_processing,
                    current_count,
                    start_index,
                    last_block_quote_index,
                ) = BlockQuoteCountHelper.__xx(
                    parser_state,
                    adjusted_line,
                    start_index,
                    current_count,
                    stack_count,
                    last_block_quote_index,
                )
            else:
                continue_processing = True
        return (
            continue_processing,
            avoid_block_starts,
            start_index,
            adjusted_line,
            last_block_quote_index,
            current_count,
        )

    @staticmethod
    def __xx_part_one(
        parser_state: ParserState, start_index: int, current_count: int
    ) -> Tuple[bool, int]:
        if parser_state.token_stack[-1].is_fenced_code_block:
            return False, -1
        assert parser_state.original_line_to_parse is not None
        block_quote_character_count = ParserHelper.count_characters_in_text(
            parser_state.original_line_to_parse[:start_index], ">"
        )
        if block_quote_character_count > current_count:
            return False, -1
        count_block_quotes = 0
        found_index = -2
        for stack_token_index, stack_token in enumerate(
            parser_state.token_stack
        ):  # pragma: no cover
            if stack_token.is_block_quote:
                count_block_quotes += 1
                if count_block_quotes == block_quote_character_count:
                    found_index = stack_token_index
                    break
        assert found_index != -2
        assert found_index != len(
            parser_state.token_stack
        ), "should have completed before this"
        found_index += 1
        return (
            not parser_state.token_stack[found_index].is_block_quote,
            found_index,
        )

    @staticmethod
    def __xx_part_two(
        parser_state: ParserState,
        stack_index: int,
        start_index: int,
        current_count: int,
        stack_count: int,
        last_block_quote_index: int,
    ) -> Tuple[int, int, int]:
        # At this point, we have a "prefix", which may be partial, that has the
        # current_count of > characters, and ends with a list. If we are here,
        # we know that previous lines have had at least one more > character and
        # counted block quote.
        assert parser_state.token_stack[
            stack_index
        ].is_list, "If not bq, must be a list."
        while parser_state.token_stack[stack_index].is_list:
            stack_index += 1
        embedded_list_stack_token = cast(
            ListStackToken, parser_state.token_stack[stack_index - 1]
        )
        assert parser_state.original_line_to_parse is not None
        if (
            embedded_list_stack_token.indent_level
            >= len(parser_state.original_line_to_parse)
            or parser_state.original_line_to_parse[
                start_index : embedded_list_stack_token.indent_level
            ].strip()
            or parser_state.original_line_to_parse[
                embedded_list_stack_token.indent_level
            ]
            != ">"
        ):
            return current_count, start_index, last_block_quote_index
        current_count += 1
        last_block_quote_index = embedded_list_stack_token.indent_level + 1

        if ParserHelper.is_character_at_index(
            parser_state.original_line_to_parse, last_block_quote_index, " "
        ):
            last_block_quote_index += 1
        if current_count < stack_count and ParserHelper.is_character_at_index(
            parser_state.original_line_to_parse, last_block_quote_index, ">"
        ):
            current_count += 1
            last_block_quote_index += 1
            if ParserHelper.is_character_at_index(
                parser_state.original_line_to_parse, last_block_quote_index, " "
            ):
                last_block_quote_index += 1

        return current_count, last_block_quote_index, last_block_quote_index

    # pylint: enable=too-many-arguments
    @staticmethod
    def __is_special_double_block_case(
        parser_state: ParserState,
        adjusted_line: str,
        start_index: int,
        current_count: int,
        stack_count: int,
    ) -> Tuple[bool, int]:
        continue_processing = False
        POGGER.debug("not block>$ of :$:", start_index, adjusted_line)
        POGGER.debug("not block>:$:", adjusted_line[start_index:])
        if current_count < stack_count:
            final_stack_index = BlockQuoteCountHelper.__find_double_block_case_index(
                parser_state, current_count
            )
            POGGER.debug(
                ">>stack>:$:$:",
                final_stack_index,
                parser_state.token_stack[final_stack_index],
            )
            POGGER.debug(
                "+1>>stack>:$:$:",
                final_stack_index + 1,
                parser_state.token_stack[final_stack_index + 1],
            )
            if parser_state.token_stack[final_stack_index + 1].is_block_quote:
                next_bq_index = adjusted_line.find(
                    BlockQuoteCountHelper.__block_quote_character, start_index
                )
                POGGER.debug("+1>>next_bq_index:$:", next_bq_index)
                index_delta = next_bq_index - start_index
                if next_bq_index != -1 and index_delta <= 3:
                    final_index, _ = ParserHelper.collect_while_spaces(
                        adjusted_line, start_index
                    )
                    if final_index == next_bq_index:
                        continue_processing, start_index = True, next_bq_index
                    else:
                        cast(
                            BlockQuoteMarkdownToken,
                            parser_state.token_stack[
                                final_stack_index
                            ].matching_markdown_token,
                        ).weird_kludge_seven = True
        return continue_processing, start_index

    @staticmethod
    def __find_double_block_case_index(
        parser_state: ParserState, current_count: int
    ) -> int:
        count_to_consume, stack_index, final_stack_index = current_count, 0, 0
        while not final_stack_index and stack_index < len(parser_state.token_stack):
            stack_token = parser_state.token_stack[stack_index]
            POGGER.debug("stack>:$:$:", stack_index, stack_token)
            if stack_token.is_block_quote:
                count_to_consume -= 1
                if not count_to_consume:
                    final_stack_index = stack_index
            stack_index += 1
        assert (
            not count_to_consume and final_stack_index
        ), "We expected to find something, so verify that we did."
        return final_stack_index

    # pylint: disable=too-many-arguments
    @staticmethod
    def __increase_stack(
        parser_state: ParserState,
        container_level_tokens: List[MarkdownToken],
        block_quote_data: BlockQuoteData,
        position_marker: PositionMarker,
        original_start_index: int,
        container_start_bq_count: int,
        extracted_whitespace: str,
    ) -> Tuple[int, BlockQuoteData]:
        POGGER.debug("container_level_tokens>>$", container_level_tokens)
        POGGER.debug("current_count>>$", block_quote_data.current_count)
        POGGER.debug("stack_count>>$", block_quote_data.stack_count)
        stack_count = block_quote_data.stack_count
        while block_quote_data.current_count > stack_count:
            POGGER.debug(
                "increasing block quotes by one>>",
            )
            stack_count += 1

            adjusted_position_marker = PositionMarker(
                position_marker.line_number,
                original_start_index,
                position_marker.text_to_parse,
            )

            if container_start_bq_count:
                POGGER.debug("extracted_whitespace>>$<<", extracted_whitespace)
                POGGER.debug("container_start_bq_count>>$<<", container_start_bq_count)
                POGGER.debug("original_start_index>>$<<", original_start_index)
                extracted_whitespace = extracted_whitespace[original_start_index:]
                POGGER.debug("extracted_whitespace>>$<<", extracted_whitespace)

            assert (
                position_marker.text_to_parse[original_start_index]
                == BlockQuoteCountHelper.__block_quote_character
            ), "Character at index must be a > character to start the block quote."
            original_start_index += 1
            if ParserHelper.is_character_at_index_whitespace(
                position_marker.text_to_parse, original_start_index
            ):
                original_start_index += 1

            new_markdown_token = BlockQuoteMarkdownToken(
                extracted_whitespace, adjusted_position_marker
            )

            container_level_tokens.append(new_markdown_token)
            parser_state.token_stack.append(BlockQuoteStackToken(new_markdown_token))

        POGGER.debug("container_level_tokens>>$", container_level_tokens)
        skip = BlockQuoteCountHelper.__block_list_block_kludge(
            parser_state,
            position_marker,
            stack_count,
            block_quote_data,
        )
        if not skip:
            block_quote_data = BlockQuoteCountHelper.decrease_stack_to_level(
                parser_state,
                block_quote_data.current_count,
                stack_count,
                container_level_tokens,
            )
        POGGER.debug(
            "container_level_tokens>>$",
            container_level_tokens,
        )

        return original_start_index, block_quote_data

    # pylint: enable=too-many-arguments

    @staticmethod
    def decrease_stack_to_level(
        parser_state: ParserState,
        current_count: int,
        stack_count: int,
        container_level_tokens: List[MarkdownToken],
    ) -> BlockQuoteData:
        """
        Decrease the stack to match the block quote level.
        """
        while current_count < stack_count:
            POGGER.debug(
                "decreasing block quotes by one>>",
            )
            stack_count -= 1
            (
                new_tokens,
                _,
            ) = parser_state.close_open_blocks_fn(
                parser_state,
                include_block_quotes=True,
                until_this_index=len(parser_state.token_stack) - 1,
                was_forced=True,
            )
            container_level_tokens.extend(new_tokens)
        return BlockQuoteData(current_count, stack_count)

    @staticmethod
    def __block_list_block_kludge(
        parser_state: ParserState,
        position_marker: PositionMarker,
        stack_count: int,
        block_quote_data: BlockQuoteData,
    ) -> bool:
        POGGER.debug("original_line_to_parse>>$", parser_state.original_line_to_parse)
        POGGER.debug("text_to_parse>>$", position_marker.text_to_parse)
        POGGER.debug("index_number>>$", position_marker.index_number)
        POGGER.debug("index_indent>>$", position_marker.index_indent)

        POGGER.debug("stack_count>>$", stack_count)
        POGGER.debug("block_quote_data.stack_count>>$", block_quote_data.stack_count)
        POGGER.debug(
            "block_quote_data.current_count>>$", block_quote_data.current_count
        )

        # KLUDGE!
        skip = False
        if (
            parser_state.original_line_to_parse == position_marker.text_to_parse
            and not position_marker.index_indent
        ):
            last_active_block_quote_stack_index = position_marker.text_to_parse[
                : position_marker.index_number + 1
            ].count(">")
            POGGER.debug(
                "last_active_block_quote_stack_index>>$",
                last_active_block_quote_stack_index,
            )

            text_after_current_block_quote = position_marker.text_to_parse[
                position_marker.index_number + 1 :
            ]
            found_index = next(
                (
                    char_index
                    for char_index, current_char in enumerate(
                        text_after_current_block_quote
                    )
                    if current_char not in " >"
                ),
                -1,
            )
            POGGER.debug("found_index:$", found_index)
            assert found_index < len(
                text_after_current_block_quote
            ), "Found index must be within string."
            whitespace_after_block_quote = text_after_current_block_quote[:found_index]
            text_after_block_quote = text_after_current_block_quote[found_index:]
            POGGER.debug(
                "whitespace_after_block_quote:$: + text_after_block_quote:$:",
                whitespace_after_block_quote,
                text_after_block_quote,
            )

            POGGER.debug("token_stack>>$", parser_state.token_stack)
            POGGER.debug(
                "token_stack[last_active_block_quote_stack_index]>>$",
                parser_state.token_stack[last_active_block_quote_stack_index],
            )
            stack_index_valid = last_active_block_quote_stack_index + 1 < len(
                parser_state.token_stack
            )
            stack_index_in_scope = last_active_block_quote_stack_index < stack_count
            more_block_quotes_present = ">" in whitespace_after_block_quote
            POGGER.debug(
                "stack_index_valid:$ and stack_index_in_scope:$ and more_block_quotes_present:$",
                stack_index_valid,
                stack_index_in_scope,
                more_block_quotes_present,
            )
            if stack_index_valid and stack_index_in_scope and more_block_quotes_present:
                POGGER.debug(
                    "xy>>$",
                    parser_state.token_stack[last_active_block_quote_stack_index + 1],
                )
                skip = parser_state.token_stack[
                    last_active_block_quote_stack_index + 1
                ].is_list
        return skip

    @staticmethod
    def __does_require_increase_or_descrease(
        parser_state: ParserState, block_quote_data: BlockQuoteData
    ) -> Tuple[bool, bool]:
        POGGER.debug(
            "__ensure_stack_at_level>>block_quote_data.current_count>>$>>block_quote_data.stack_count>>$",
            block_quote_data.current_count,
            block_quote_data.stack_count,
        )
        stack_increase_needed = (
            block_quote_data.current_count > block_quote_data.stack_count
        )
        if (
            not stack_increase_needed
            and block_quote_data.current_count < block_quote_data.stack_count
        ):
            POGGER.debug(
                "__ensure_stack_at_level>>possible decrease to new level",
            )
            top_token_on_stack = parser_state.token_stack[-1]
            POGGER.debug("__ensure_stack_at_level>>$", top_token_on_stack)
            stack_decrease_needed = (
                top_token_on_stack.is_indented_code_block
                or top_token_on_stack.is_html_block
                or top_token_on_stack.is_list
                or top_token_on_stack.is_block_quote
            )
            POGGER.debug(
                "__ensure_stack_at_level>>decrease to new level=$",
                stack_decrease_needed,
            )
        else:
            stack_decrease_needed = False
        return stack_increase_needed, stack_decrease_needed

    @staticmethod
    def __decrease_stack(
        parser_state: ParserState,
        container_level_tokens: List[MarkdownToken],
        original_start_index: int,
        stack_hard_limit: Optional[int],
    ) -> None:
        POGGER.debug("token_stack>>$", parser_state.token_stack)
        POGGER.debug("token_document>>$", parser_state.token_document)
        POGGER.debug(
            "container_level_tokens>>$",
            container_level_tokens,
        )
        POGGER.debug("stack_hard_limit>>$", stack_hard_limit)
        stack_conditional = stack_hard_limit is None or (
            len(parser_state.token_stack) > stack_hard_limit
        )
        POGGER.debug("stack_conditional>>$", stack_conditional)
        while stack_conditional and parser_state.token_stack[-1].is_list:
            list_stack_token = cast(ListStackToken, parser_state.token_stack[-1])
            POGGER.debug("stack>>$", list_stack_token.indent_level)
            POGGER.debug("original_start_index>>$", original_start_index)
            if original_start_index < list_stack_token.indent_level:
                close_tokens, _ = parser_state.close_open_blocks_fn(
                    parser_state,
                    include_lists=True,
                    was_forced=True,
                    until_this_index=len(parser_state.token_stack) - 1,
                )
                container_level_tokens.extend(close_tokens)
                POGGER.debug("container_level_tokens>>$", container_level_tokens)
            else:
                break  # pragma: no cover

    # pylint: disable=too-many-arguments
    @staticmethod
    def ensure_stack_at_level(
        parser_state: ParserState,
        block_quote_data: BlockQuoteData,
        extracted_whitespace: str,
        position_marker: PositionMarker,
        original_start_index: int,
        container_start_bq_count: int,
    ) -> Tuple[
        List[MarkdownToken],
        Optional[RequeueLineInfo],
        bool,
        BlockQuoteData,
    ]:
        """
        Ensure that the block quote stack is at the proper level on the stack.
        """
        container_level_tokens: List[MarkdownToken] = []
        (
            stack_increase_needed,
            stack_decrease_needed,
        ) = BlockQuoteCountHelper.__does_require_increase_or_descrease(
            parser_state, block_quote_data
        )

        POGGER.debug("current_count>>$", block_quote_data.current_count)
        POGGER.debug("stack_count>>$", block_quote_data.stack_count)
        POGGER.debug(
            "stack_increase_needed>>$, stack_decrease_needed=$",
            stack_increase_needed,
            stack_decrease_needed,
        )
        if stack_increase_needed or stack_decrease_needed:
            POGGER.debug(
                "token_stack>>$",
                parser_state.token_stack,
            )
            POGGER.debug("token_document>>$", parser_state.token_document)
            (
                container_level_tokens,
                requeue_line_info,
            ) = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[
                    ParagraphStackToken,
                    IndentedCodeBlockStackToken,
                    LinkDefinitionStackToken,
                    HtmlBlockStackToken,
                    TableBlockStackToken,
                ],
                was_forced=True,
                caller_can_handle_requeue=True,
                requeue_reset=True,
            )
            if requeue_line_info:
                return [], requeue_line_info, False, block_quote_data

            POGGER.debug("esal>>__calculate_stack_hard_limit(delta)")
            (
                stack_hard_limit,
                force_list_continuation,
            ) = BlockQuoteCountHelper.__calculate_stack_hard_limit(
                parser_state,
                position_marker,
                False,
                stack_increase_needed,
                stack_decrease_needed,
                block_quote_data,
            )
            POGGER.debug("force_list_continuation=$", force_list_continuation)
            POGGER.debug("esal>>__calculate_stack_hard_limit>>$", stack_hard_limit)

            BlockQuoteCountHelper.__decrease_stack(
                parser_state,
                container_level_tokens,
                original_start_index,
                stack_hard_limit,
            )

            (
                original_start_index,
                block_quote_data,
            ) = BlockQuoteCountHelper.__increase_stack(
                parser_state,
                container_level_tokens,
                block_quote_data,
                position_marker,
                original_start_index,
                container_start_bq_count,
                extracted_whitespace,
            )
        else:
            POGGER.debug("esal>>__calculate_stack_hard_limit(no delta)")
            (
                stack_hard_limit,
                force_list_continuation,
            ) = BlockQuoteCountHelper.__calculate_stack_hard_limit(
                parser_state, position_marker, True, False, False, block_quote_data
            )
            POGGER.debug("esal>>__calculate_stack_hard_limit>>$", stack_hard_limit)

        return (
            container_level_tokens,
            None,
            force_list_continuation,
            block_quote_data,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_stack_hard_limit(
        parser_state: ParserState,
        position_marker: PositionMarker,
        adjust_current_block_quote: bool,
        stack_increase_needed: bool,
        stack_decrease_needed: bool,
        block_quote_data: BlockQuoteData,
    ) -> Tuple[Optional[int], bool]:
        POGGER.debug(">>__calculate_stack_hard_limit>>")
        POGGER.debug("original_line_to_parse>>:$:", parser_state.original_line_to_parse)
        POGGER.debug(
            "position_marker>>[$:$]:$:",
            position_marker.index_indent,
            position_marker.index_number,
            position_marker.text_to_parse,
        )

        stack_hard_limit, extra_consumed_whitespace, last_bq_index = (
            None,
            None,
            parser_state.find_last_block_quote_on_stack(),
        )

        assert (
            parser_state.original_line_to_parse is not None
        ), "The original line to parse must be defined."
        # TDone need to find a better way, stopgap
        # conditional_1 = parser_state.original_line_to_parse.endswith(
        #     position_marker.text_to_parse
        # )
        # POGGER.debug(
        #     "conditional_1:$: = oltp:$:endswith(ttp:$:)",
        #     conditional_1,
        #     parser_state.original_line_to_parse,
        #     position_marker.text_to_parse,
        # )
        # conditional_2 = (
        #     len(parser_state.token_stack) > 2
        #     and parser_state.token_stack[1].is_block_quote
        # )
        # POGGER.debug(
        #     "conditional_2:$: = len(ts:$:) > 2 and ts[1].is_bq:$:",
        #     conditional_2,
        #     parser_state.token_stack,
        #     (
        #         parser_state.token_stack[1].is_block_quote
        #         if len(parser_state.token_stack) > 2
        #         else None
        #     ),
        # )
        # conditional_3 = last_bq_index != 1 or stack_increase_needed
        # POGGER.debug(
        #     "conditional_3:$: = lbl:$: != 1 or stack_increase_needed:$:",
        #     conditional_3,
        #     last_bq_index,
        #     stack_increase_needed,
        # )
        # POGGER.debug(
        #     "conditional_1>>:$ and ty2:$: and ty3:$:",
        #     conditional_1,
        #     conditional_2,
        #     conditional_3,
        # )
        # if conditional_1 and conditional_2 and conditional_3:
        if (
            parser_state.original_line_to_parse.endswith(position_marker.text_to_parse)
            and (
                len(parser_state.token_stack) > 2
                and parser_state.token_stack[1].is_block_quote
            )
            and (last_bq_index != 1 or stack_increase_needed)
        ):
            (
                stack_hard_limit,
                extra_consumed_whitespace,
                force_list_continuation,
            ) = BlockQuoteCountHelper.__calculate_limit_and_continuation(
                parser_state,
                position_marker,
                adjust_current_block_quote,
                last_bq_index,
                stack_increase_needed,
                stack_decrease_needed,
                block_quote_data,
            )
        else:
            force_list_continuation = False

        POGGER.debug(
            "<<__calculate_stack_hard_limit<<$,$",
            stack_hard_limit,
            extra_consumed_whitespace,
        )
        POGGER.debug("force_list_continuation=$", force_list_continuation)
        return stack_hard_limit, force_list_continuation

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_limit_and_continuation(
        parser_state: ParserState,
        position_marker: PositionMarker,
        adjust_current_block_quote: bool,
        last_bq_index: int,
        stack_increase_needed: bool,
        stack_decrease_needed: bool,
        block_quote_data: BlockQuoteData,
    ) -> Tuple[Optional[int], Optional[int], bool]:
        (
            length_of_available_whitespace,
            _,
        ) = ParserHelper.extract_spaces_verified(position_marker.text_to_parse, 0)
        POGGER.debug("len(ws)>>:$:", length_of_available_whitespace)

        (
            stack_hard_limit,
            extra_consumed_whitespace,
        ) = BlockQuoteCountHelper.__calculate_stack_hard_limit_if_eligible(
            parser_state,
            position_marker,
            # length_of_available_whitespace,
            # adjust_current_block_quote,
            # last_bq_index,
        )
        POGGER.debug(f">>>>>stack_increase_needed:{stack_increase_needed}")
        POGGER.debug(f">>>>>stack_decrease_needed:{stack_decrease_needed}")
        POGGER.debug(f">>>>>adjust_current_block_quote:{adjust_current_block_quote}")
        POGGER.debug(
            ">>block_quote_data.current_count>>$",
            block_quote_data.current_count,
        )
        POGGER.debug(">>block_quote_data.stack_count>>$", block_quote_data.stack_count)
        force_list_continuation = (
            extra_consumed_whitespace is None
            and not stack_increase_needed
            and not stack_decrease_needed
            and adjust_current_block_quote
        )
        if force_list_continuation:
            POGGER.debug(f">>>>>last_bq_index:{last_bq_index}")
            POGGER.debug(f">>>>>parser_state.token_stack:{parser_state.token_stack}")
            POGGER.debug(
                f">>>>>len(parser_state.token_stack):{len(parser_state.token_stack)}"
            )
            force_list_continuation = (
                last_bq_index + 1 < len(parser_state.token_stack)
                and parser_state.token_stack[last_bq_index + 1].is_list
                and block_quote_data.current_count != block_quote_data.stack_count
            )
        return stack_hard_limit, extra_consumed_whitespace, force_list_continuation

    # pylint: enable=too-many-arguments

    @staticmethod
    def __calculate_stack_hard_limit_if_eligible(
        parser_state: ParserState,
        position_marker: PositionMarker,
        # length_of_available_whitespace: int,
        # adjust_current_block_quote: bool,
        # last_bq_index: int,
    ) -> Tuple[Optional[int], Optional[int]]:
        POGGER.debug("eligible")
        stack_hard_limit, extra_consumed_whitespace = None, None
        assert (
            parser_state.original_line_to_parse is not None
        ), "The original line to parse must be defined."
        remaining_text = parser_state.original_line_to_parse[
            : -len(position_marker.text_to_parse)
        ]
        assert not remaining_text
        # POGGER.debug("eligible - remaining_text:$:", remaining_text)

        # # use up already extracted text/ws
        # current_stack_index, indent_text_count, extra_consumed_whitespace = 1, 0, 0
        # assert parser_state.token_stack[
        #     current_stack_index
        # ].is_block_quote, "The specified index must point to a block quote."
        # POGGER.debug("bq")
        # start_index = remaining_text.find(">")
        # assert (
        #     start_index != -1
        # ), "A block quote start character (>) must be present in the remaing text to parse."
        # POGGER.debug("bq-found")
        # indent_text_count = start_index + 1
        # assert (
        #     indent_text_count < len(remaining_text)
        #     and remaining_text[indent_text_count] == ParserHelper.space_character
        # ), "The indent_text_count needs rooms to grow, and currently specifies a space character."
        # POGGER.debug("bq-space-found")
        # indent_text_count += 1
        # current_stack_index += 1
        # assert indent_text_count == len(
        #     remaining_text
        # ), "Counts should equal each other."

        # # if there is whitespace
        # (
        #     stack_hard_limit,
        #     indent_text_count,
        #     length_of_available_whitespace,
        #     extra_consumed_whitespace,
        # ) = BlockQuoteCountHelper.__calculate_eligible_stack_hard_limit(
        #     parser_state,
        #     current_stack_index,
        #     indent_text_count,
        #     length_of_available_whitespace,
        #     extra_consumed_whitespace,
        #     adjust_current_block_quote,
        #     last_bq_index,
        # )
        return stack_hard_limit, extra_consumed_whitespace

    # # pylint: disable=too-many-arguments
    # @staticmethod
    # def __calculate_eligible_stack_hard_limit(
    #     parser_state: ParserState,
    #     current_stack_index: int,
    #     indent_text_count: int,
    #     length_of_available_whitespace: int,
    #     extra_consumed_whitespace: int,
    #     adjust_current_block_quote: bool,
    #     last_bq_index: int,
    # ) -> Tuple[int, int, int, int]:
    #     assert parser_state.token_stack[
    #         current_stack_index
    #     ].is_list, "Since we have taken care of any block quotes, this must be a list."
    #     list_stack_token = cast(
    #         ListStackToken, parser_state.token_stack[current_stack_index]
    #     )
    #     POGGER.debug(
    #         "indent_level:$:indent_text_count:$:",
    #         list_stack_token.indent_level,
    #         indent_text_count,
    #     )
    #     delta = list_stack_token.indent_level - indent_text_count
    #     POGGER.debug(
    #         "delta:$:length_of_available_whitespace:$:",
    #         delta,
    #         length_of_available_whitespace,
    #     )
    #     assert (
    #         length_of_available_whitespace >= delta
    #     ), "Amount of whitespace to start with should always be greater than the amount to reduce by."
    #     list_token = cast(
    #         ListStartMarkdownToken, list_stack_token.matching_markdown_token
    #     )
    #     adjust_for_extra_indent = list_token.indent_level - list_token.column_number - 1
    #     if list_stack_token.is_ordered_list:
    #         adjust_for_extra_indent -= len(list_token.list_start_sequence) - 1
    #     POGGER.debug("adjust_for_extra_indent:$:", adjust_for_extra_indent)
    #     current_stack_index += 1
    #     delta -= adjust_for_extra_indent
    #     indent_text_count += delta
    #     length_of_available_whitespace -= delta
    #     extra_consumed_whitespace += delta

    #     assert not adjust_current_block_quote
    #     _ = last_bq_index
    #     # if adjust_current_block_quote:
    #     #     POGGER.debug(
    #     #         "__calculate_stack_hard_limit>>last_block_token>>$",
    #     #         parser_state.token_stack[last_bq_index].matching_markdown_token,
    #     #     )
    #     #     block_token = cast(
    #     #         BlockQuoteMarkdownToken,
    #     #         parser_state.token_stack[last_bq_index].matching_markdown_token,
    #     #     )
    #     #     block_token.add_bleading_spaces(
    #     #         ParserHelper.repeat_string(ParserHelper.space_character, delta), True
    #     #     )
    #     #     POGGER.debug(
    #     #         "__calculate_stack_hard_limit>>last_block_token>>$", block_token
    #     #     )

    #     return (
    #         current_stack_index,
    #         indent_text_count,
    #         length_of_available_whitespace,
    #         extra_consumed_whitespace,
    #     )

    # # pylint: enable=too-many-arguments



"""
Module to help with the parsing of character reference inline elements.
"""

import json
import logging
import os
import string
from typing import Dict, Optional, Tuple

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.bad_tokenization_error import BadTokenizationError
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.inline_request import InlineRequest
from pymarkdown.inline.inline_response import InlineResponse

POGGER = ParserLogger(logging.getLogger(__name__))


class InlineCharacterReferenceHelper:
    """
    Class to help with the parsing of character reference inline elements.
    """

    __invalid_reference_character_substitute = "\ufffd"

    __hex_character_reference_start_character = "xX"
    __numeric_character_reference_start_character = "#"

    character_reference_start_character = "&"
    __character_reference_end_character = ";"

    __skip_html5_entities_ending_with = ";"

    __ascii_letters_and_digits = f"{string.ascii_letters}{string.digits}"

    __entity_map: Dict[str, str] = {}
    __entities_file_name = "entities.json"

    @staticmethod
    def initialize(resource_path: str) -> None:
        """
        Initialize the inline subsystem.
        """
        InlineCharacterReferenceHelper.__entity_map = (
            InlineCharacterReferenceHelper.__load_entity_map(resource_path)
        )

    @staticmethod
    def handle_character_reference(
        parser_properties: ParseBlockPassProperties, inline_request: InlineRequest
    ) -> InlineResponse:
        """
        Handle a generic character reference.
        """
        inline_response = InlineResponse()
        inline_response.new_index, inline_response.new_string, source_text_size = (
            inline_request.next_index + 1,
            "",
            len(inline_request.source_text),
        )
        assert (
            inline_response.new_index is not None
        ), "new_index should be defined by this point."
        if (
            inline_response.new_index < source_text_size
            and inline_request.source_text[inline_response.new_index]
            == InlineCharacterReferenceHelper.__numeric_character_reference_start_character
        ):
            InlineCharacterReferenceHelper.__handle_numeric_character_reference(
                inline_request, inline_response
            )
        else:
            InlineCharacterReferenceHelper.__handle_non_numeric_character_reference(
                inline_request, inline_response, source_text_size
            )
        inline_response.delta_line_number, inline_response.delta_column_number = (
            0,
            inline_response.new_index - inline_request.next_index,
        )
        return inline_response

    @staticmethod
    def __load_entity_map(resource_path: str) -> Dict[str, str]:
        """
        Load the entity map, refreshed from https://html.spec.whatwg.org/entities.json
        into a dict that was can use.
        """

        master_entities_file = os.path.join(
            resource_path, InlineCharacterReferenceHelper.__entities_file_name
        )
        try:
            with open(
                os.path.abspath(master_entities_file), "rt", encoding="utf-8"
            ) as infile:
                results_dictionary = json.load(infile)
        except json.decoder.JSONDecodeError as this_exception:
            error_message = (
                f"Named character entity map file '{master_entities_file}' "
                + f"is not a valid JSON file ({this_exception})."
            )
            raise BadTokenizationError(error_message) from this_exception
        except IOError as this_exception:
            error_message = (
                f"Named character entity map file '{master_entities_file}' "
                + f"was not loaded ({this_exception})."
            )
            raise BadTokenizationError(error_message) from this_exception

        approved_entity_map = {}
        for next_name, char_entity in results_dictionary.items():
            # Downloaded file is for HTML5, which includes some names that do
            # not end with ";".  These are excluded.
            if (
                next_name[-1]
                != InlineCharacterReferenceHelper.__skip_html5_entities_ending_with
            ):
                continue

            entity_characters = char_entity["characters"]
            entity_codepoints = char_entity["codepoints"]

            # The only entities we should encounter either have a length of 1 or 2
            if len(entity_characters) == 1:
                assert len(entity_codepoints) == 1, "Codepoint validation."
            else:
                assert len(entity_codepoints) == 2, "Codepoint validation."
                assert (
                    ord(entity_characters[1]) == entity_codepoints[1]
                ), "Codepoint validation."
            assert (
                ord(entity_characters[0]) == entity_codepoints[0]
            ), "Codepoint validation."
            approved_entity_map[next_name] = entity_characters
        return approved_entity_map

    @staticmethod
    def __handle_numeric_character_reference_hex(
        new_index: int, source_text: str, source_text_size: int
    ) -> Tuple[str, int, int]:
        hex_char = source_text[new_index]
        new_index += 1
        end_index, collected_string = (
            ParserHelper.collect_while_one_of_characters_verified(
                source_text, new_index, string.hexdigits
            )
        )
        POGGER.debug(
            "&#x>>a>>$>>b>>$>>$", end_index, collected_string, source_text_size
        )

        delta = end_index - new_index
        POGGER.debug("delta>>$>>", delta)
        translated_reference = int(collected_string, 16) if 1 <= delta <= 6 else -1
        new_string, new_index = (
            f"{InlineCharacterReferenceHelper.character_reference_start_character}"
            + f"{InlineCharacterReferenceHelper.__numeric_character_reference_start_character}{hex_char}{collected_string}"
        ), end_index

        return new_string, new_index, translated_reference

    @staticmethod
    def __handle_numeric_character_reference_decimal(
        new_index: int, source_text: str, source_text_size: int
    ) -> Tuple[str, int, int]:
        end_index, collected_string = (
            ParserHelper.collect_while_one_of_characters_verified(
                source_text, new_index, string.digits
            )
        )
        POGGER.debug("&#>>a>>$>>b>>$>>$", end_index, collected_string, source_text_size)

        delta = end_index - new_index
        POGGER.debug("delta>>$>>", delta)
        translated_reference = int(collected_string) if 1 <= delta <= 7 else -1
        new_string, new_index = (
            f"{InlineCharacterReferenceHelper.character_reference_start_character}"
            + f"{InlineCharacterReferenceHelper.__numeric_character_reference_start_character}{collected_string}"
        ), end_index

        return new_string, new_index, translated_reference

    @staticmethod
    def __handle_numeric_character_reference_inner(
        source_text: str, new_index: int
    ) -> Tuple[str, int, Optional[str]]:
        """
        Handle a character reference that is numeric in nature.
        """

        original_reference, new_index, source_text_size = (
            None,
            new_index + 1,
            len(source_text),
        )
        if new_index < source_text_size and (
            source_text[new_index]
            in InlineCharacterReferenceHelper.__hex_character_reference_start_character
        ):
            (
                new_string,
                new_index,
                translated_reference,
            ) = InlineCharacterReferenceHelper.__handle_numeric_character_reference_hex(
                new_index, source_text, source_text_size
            )
        else:
            (
                new_string,
                new_index,
                translated_reference,
            ) = InlineCharacterReferenceHelper.__handle_numeric_character_reference_decimal(
                new_index, source_text, source_text_size
            )

        if (
            translated_reference >= 0
            and new_index < source_text_size
            and source_text[new_index]
            == InlineCharacterReferenceHelper.__character_reference_end_character
        ):
            new_index += 1
            original_reference, new_string = f"{new_string};", (
                InlineCharacterReferenceHelper.__invalid_reference_character_substitute
                if translated_reference == 0
                else chr(translated_reference)
            )
        return new_string, new_index, original_reference

    @staticmethod
    def __handle_numeric_character_reference(
        inline_request: InlineRequest, inline_response: InlineResponse
    ) -> None:
        original_new_index = inline_response.new_index
        POGGER.debug("here")
        assert (
            inline_response.new_index is not None
        ), "new_index should be defined by this point."
        (
            inline_response.new_string,
            inline_response.new_index,
            inline_response.original_string,
        ) = InlineCharacterReferenceHelper.__handle_numeric_character_reference_inner(
            inline_request.source_text, inline_response.new_index
        )
        inline_response.new_string_unresolved = (
            InlineCharacterReferenceHelper.character_reference_start_character
            + inline_request.source_text[original_new_index : inline_response.new_index]
        )
        POGGER.debug("here-->$<--", inline_response.new_string)
        POGGER.debug("here-->$<--", inline_response.new_string_unresolved)

    @staticmethod
    def __handle_non_numeric_character_reference(
        inline_request: InlineRequest,
        inline_response: InlineResponse,
        source_text_size: int,
    ) -> None:
        POGGER.debug("there")
        assert (
            inline_response.new_index is not None
        ), "new_index should be defined by this point."
        end_index, collected_string = ParserHelper.collect_while_one_of_characters(
            inline_request.source_text,
            inline_response.new_index,
            InlineCharacterReferenceHelper.__ascii_letters_and_digits,
        )
        if collected_string:
            assert (
                end_index is not None
            ), "if collected_string is defined, end_index should also be defined"
            collected_string = f"{InlineCharacterReferenceHelper.character_reference_start_character}{collected_string}"
            if (
                end_index < source_text_size
                and inline_request.source_text[end_index]
                == InlineCharacterReferenceHelper.__character_reference_end_character
            ):
                end_index += 1
                collected_string += (
                    InlineCharacterReferenceHelper.__character_reference_end_character
                )
                if collected_string in InlineCharacterReferenceHelper.__entity_map:
                    inline_response.new_string_unresolved = collected_string
                    inline_response.original_string = collected_string
                    collected_string = InlineCharacterReferenceHelper.__entity_map[
                        collected_string
                    ]
            inline_response.new_string, inline_response.new_index = (
                collected_string,
                end_index,
            )
            POGGER.debug("there-->$<--", inline_response.new_string)
            POGGER.debug("there-->$<--", inline_response.new_string_unresolved)
        else:
            inline_response.new_string = (
                InlineCharacterReferenceHelper.character_reference_start_character
            )



"""
Module to implement a plugin that looks for text sequences that make
it appear like the author got the inline link syntax reversed.
"""

import re
from typing import List, Optional

from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd011(RulePlugin):
    """
    Class to implement a plugin that looks for text sequences that make
    it appear like the author got the inline link syntax reversed.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__reverse_link_syntax = re.compile(r"\(.*\)\[\s*[^\^].*\s*]")
        self.__leaf_tokens: List[MarkdownToken] = []
        self.__line_index = 0
        self.__leaf_token_index = 0
        self.__last_line: Optional[str] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-reversed-links",
            plugin_id="MD011",
            plugin_enabled_by_default=True,
            plugin_description="Reversed link syntax",
            plugin_version="0.5.1",
            plugin_interface_version=1,
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md011.md",
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__leaf_tokens = []
        self.__line_index = 1
        self.__leaf_token_index = 0
        self.__last_line = None

    # pylint: disable=too-many-boolean-expressions
    def next_line(self, context: PluginScanContext, line: str) -> None:
        """
        Event that a new line is being processed.
        """
        if (
            self.__leaf_token_index + 1 < len(self.__leaf_tokens)
            and self.__line_index
            == self.__leaf_tokens[self.__leaf_token_index + 1].line_number
        ):
            self.__leaf_token_index += 1

        if (
            self.__leaf_token_index < len(self.__leaf_tokens)
            and not self.__leaf_tokens[self.__leaf_token_index].is_code_block
            and not self.__leaf_tokens[self.__leaf_token_index].is_html_block
            and line
            and "(" in line
            and "[" in line
        ):
            if regex_search := self.__reverse_link_syntax.search(line):
                regex_span = regex_search.span()
                extra_error_information = line[regex_span[0] : regex_span[1]]
                override_is_error_token_prefaced_by_blank_line = (
                    self.__last_line is not None and not self.__last_line
                )
                self.report_next_line_error(
                    context,
                    regex_span[0] + 1,
                    extra_error_information=extra_error_information,
                    override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
                )

        self.__line_index += 1
        self.__last_line = line.strip(" \t")

    # pylint: enable=too-many-boolean-expressions

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        _ = context
        if token.is_blank_line or token.is_leaf:
            self.__leaf_tokens.append(token)



"""
Module to provide helper functions for containers.

These functions are typically needed in other modules than the two main
container modules.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.stack_token import ListStackToken

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-few-public-methods


class ContainerHelper:
    """
    Class to provide helper functions for containers.
    """

    @staticmethod
    def __reduce_containers_if_required_bq_list(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        new_tokens: List[MarkdownToken],
    ) -> Tuple[bool, str, str]:
        did_once = False
        whitespace_prefix = ""
        # list_indent_level = None
        if parser_state.token_stack[-1].is_list:
            # leading_space_length = (
            #     len(extracted_whitespace) + position_marker.index_indent
            # )
            assert parser_state.original_line_to_parse is not None
            leading_space_length = parser_state.original_line_to_parse.index(
                position_marker.text_to_parse
            ) + len(extracted_whitespace)
            search_index = len(parser_state.token_stack)
            while parser_state.token_stack[search_index - 1].is_list:
                list_token = cast(
                    ListStackToken, parser_state.token_stack[search_index - 1]
                )
                if list_token.indent_level <= leading_space_length:
                    break
                search_index -= 1
                did_once = True

            if did_once:
                last_token = cast(EndMarkdownToken, new_tokens[-1])
                last_token.set_extra_end_data(None)

            (
                container_level_tokens,
                _,
            ) = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=search_index,
                include_lists=True,
                caller_can_handle_requeue=False,
                requeue_reset=True,
            )
            new_tokens.extend(container_level_tokens)

            indent_delta = list_token.indent_level - position_marker.index_indent
            whitespace_prefix = extracted_whitespace[:indent_delta]
            extracted_whitespace = extracted_whitespace[indent_delta:]
        return did_once, extracted_whitespace, whitespace_prefix

    @staticmethod
    def __xx(parser_state: ParserState, extra_bqs: int) -> List[MarkdownToken]:
        until_index = len(parser_state.token_stack) - 1
        needed_bqs = extra_bqs
        while until_index > 0 and needed_bqs > 0:
            if parser_state.token_stack[until_index].is_block_quote:
                needed_bqs -= 1
            until_index -= 1
        x_tokens, _ = parser_state.close_open_blocks_fn(
            parser_state,
            include_block_quotes=True,
            include_lists=True,
            was_forced=True,
            until_this_index=until_index + 1,
        )
        POGGER.debug("x_tokens>>:$:<", x_tokens)
        return x_tokens

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __reduce_containers_if_required_bq(
        parser_state: ParserState,
        position_marker: PositionMarker,
        new_tokens: List[MarkdownToken],
        split_tab: bool,
        extracted_whitespace: str,
        grab_bag: ContainerGrabBag,
        extra_bqs: int,
    ) -> Tuple[bool, str, Optional[str]]:
        x_tokens = ContainerHelper.__xx(parser_state, extra_bqs)
        count_block_quotes = sum(bool(i.is_block_quote_end) for i in x_tokens)
        assert (
            count_block_quotes == extra_bqs
        ), "Should have generated the requested number of block quote tokens."

        first_new_token = cast(EndMarkdownToken, x_tokens[0])

        did_reduce_list, extracted_whitespace, whitespace_prefix = (
            ContainerHelper.__reduce_containers_if_required_bq_list(
                parser_state, position_marker, extracted_whitespace, x_tokens
            )
        )
        was_list_ended = (
            grab_bag.container_tokens[-1].is_list_end
            if grab_bag.container_tokens and grab_bag.container_tokens[-1].is_end_token
            else False
        )

        matching_start_token = cast(
            BlockQuoteMarkdownToken, first_new_token.start_markdown_token
        )
        POGGER.debug(
            "start_markdown_token.bleading>>:$:<",
            matching_start_token.bleading_spaces,
        )
        assert (
            matching_start_token.bleading_spaces is not None
        ), "Bleading spaces must be defined by this point."
        last_newline_index = matching_start_token.bleading_spaces.rfind("\n")
        # if last_newline_index == -1:
        #     last_newline_part =matching_start_token.leading_spaces
        # else:
        last_newline_part = matching_start_token.bleading_spaces[
            last_newline_index + 1 :
        ]
        # POGGER.debug("last_newline_part>>:$:<", last_newline_part)
        if split_tab:
            # if last_newline_part.endswith(                " "            ):
            #     last_newline_part = last_newline_part[:-1]
            # POGGER.debug("last_newline_part>>:$:<", last_newline_part)
            split_tab = False
        # POGGER.debug("split_tab>>:$:<", split_tab)
        last_newline_part += whitespace_prefix

        # POGGER.debug("extra_end_data>>:$:<", first_new_token.extra_end_data)
        assert (
            first_new_token.extra_end_data is None
        ), "Extra data must be defined by this point."

        was_paragraph_closed = (
            new_tokens[0].is_paragraph_end
            if new_tokens and new_tokens[0].is_end_token
            else False
        )
        if did_reduce_list or was_list_ended or not was_paragraph_closed:
            first_new_token.set_extra_end_data(None)
        else:
            first_new_token.set_extra_end_data(last_newline_part)

        new_tokens.extend(x_tokens)
        new_whitespace_prefix = ContainerHelper.__handle_whitespace_prefix(
            parser_state, whitespace_prefix, last_newline_part
        )
        return split_tab, extracted_whitespace, new_whitespace_prefix

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __handle_whitespace_prefix(
        parser_state: ParserState, whitespace_prefix: str, last_newline_part: str
    ) -> Optional[str]:
        if not whitespace_prefix:
            return None

        indent_level = 0
        stack_index = len(parser_state.token_stack) - 1
        while stack_index > 0:
            if parser_state.token_stack[stack_index].is_list:
                indent_level += cast(
                    ListStartMarkdownToken,
                    parser_state.token_stack[stack_index].matching_markdown_token,
                ).indent_level
                break
            bleading_spaces = cast(
                BlockQuoteMarkdownToken,
                parser_state.token_stack[stack_index].matching_markdown_token,
            ).bleading_spaces
            assert bleading_spaces is not None
            split_bleading_spaces = bleading_spaces.split("\n")
            last_split_bleading_spaces = len(split_bleading_spaces[-1])
            indent_level += last_split_bleading_spaces
            stack_index -= 1
        return last_newline_part[indent_level:]

    # pylint: disable=too-many-arguments
    @staticmethod
    def reduce_containers_if_required(
        parser_state: ParserState,
        position_marker: PositionMarker,
        block_quote_data: BlockQuoteData,
        new_tokens: List[MarkdownToken],
        split_tab: bool,
        extracted_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[bool, str, Optional[str]]:
        """
        Given a drop in the current count of block quotes versus what is actually
        specified, reduce the containers.
        """

        POGGER.debug(
            "block_quote_data.current_count>>:$:<", block_quote_data.current_count
        )
        POGGER.debug("block_quote_data.stack_count>>:$:<", block_quote_data.stack_count)
        POGGER.debug("new_tokens>>:$:<", new_tokens)
        POGGER.debug("split_tab>>:$:<", split_tab)
        POGGER.debug("parser_state.token_stack[-1]>>:$:<", parser_state.token_stack[-1])

        # TODO While? needs to take lists into account as well
        whitespace_prefix: Optional[str] = None
        if (
            block_quote_data.current_count >= 0
            and block_quote_data.stack_count > block_quote_data.current_count
            and parser_state.token_stack[-1].is_block_quote
        ):
            split_tab, extracted_whitespace, whitespace_prefix = (
                ContainerHelper.__reduce_containers_if_required_bq(
                    parser_state,
                    position_marker,
                    new_tokens,
                    split_tab,
                    extracted_whitespace,
                    grab_bag,
                    block_quote_data.stack_count - block_quote_data.current_count,
                )
            )

        return split_tab, extracted_whitespace, whitespace_prefix

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
Module to provide for an encapsulation of the fenced code block element.
"""

import logging
from typing import Optional, Union, cast

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class FencedCodeBlockMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the fenced code block element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        fence_character: str,
        fence_count: int,
        extracted_text: str,
        pre_extracted_text: str,
        text_after_extracted_text: str,
        pre_text_after_extracted_text: str,
        extracted_whitespace: str,
        extracted_whitespace_before_info_string: str,
        position_marker: PositionMarker,
    ) -> None:
        (
            self.__extracted_text,
            self.__pre_extracted_text,
            self.__extracted_whitespace_before_info_string,
            self.__text_after_extracted_text,
            self.__pre_text_after_extracted_text,
            self.__fence_character,
            self.__fence_count,
        ) = (
            extracted_text,
            pre_extracted_text,
            extracted_whitespace_before_info_string,
            text_after_extracted_text,
            pre_text_after_extracted_text,
            fence_character,
            fence_count,
        )
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_fenced_code_block,
            "",
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
        )
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments
    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_fenced_code_block

    # pylint: enable=protected-access

    @property
    def fence_character(self) -> str:
        """
        Returns the character used for the fence.
        """
        return self.__fence_character

    @property
    def fence_count(self) -> int:
        """
        Returns the number of fence characters used for the fence.
        """
        return self.__fence_count

    @property
    def extracted_text(self) -> str:
        """
        Returns the text extracted from the info string.
        """
        return self.__extracted_text

    @property
    def pre_extracted_text(self) -> str:
        """
        Returns the text extracted from the info string.
        """
        return self.__pre_extracted_text

    @property
    def text_after_extracted_text(self) -> str:
        """
        Returns the text extracted after the info string.
        """
        return self.__text_after_extracted_text

    @property
    def pre_text_after_extracted_text(self) -> str:
        """
        Returns the text extracted after after the info string.
        """
        return self.__pre_text_after_extracted_text

    @property
    def extracted_whitespace_before_info_string(self) -> str:
        """
        Returns any whitespace that was extracted before the info string was processed.
        """
        return self.__extracted_whitespace_before_info_string

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        self._set_extra_data(
            MarkdownToken.extra_data_separator.join(
                [
                    self.__fence_character,
                    str(self.__fence_count),
                    self.__extracted_text,
                    self.__pre_extracted_text,
                    self.__text_after_extracted_text,
                    self.__pre_text_after_extracted_text,
                    self.extracted_whitespace,
                    self.__extracted_whitespace_before_info_string,
                ]
            )
        )

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            FencedCodeBlockMarkdownToken,
            FencedCodeBlockMarkdownToken.__rehydrate_fenced_code_block,
            FencedCodeBlockMarkdownToken.__rehydrate_fenced_code_block_end,
        )

    @staticmethod
    def __rehydrate_fenced_code_block(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the fenced code block from the token.
        """
        _ = previous_token

        context.block_stack.append(current_token)
        current_fenced_token = cast(FencedCodeBlockMarkdownToken, current_token)

        code_block_start_parts = [
            current_fenced_token.extracted_whitespace,
            ParserHelper.repeat_string(
                current_fenced_token.fence_character, current_fenced_token.fence_count
            ),
            current_fenced_token.extracted_whitespace_before_info_string,
            current_fenced_token.pre_extracted_text
            or current_fenced_token.extracted_text,
            current_fenced_token.pre_text_after_extracted_text
            or current_fenced_token.text_after_extracted_text,
            ParserHelper.newline_character,
        ]

        return "".join(code_block_start_parts)

    @staticmethod
    def __rehydrate_fenced_code_block_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:  # sourcery skip: extract-method
        """
        Rehydrate the end of the fenced code block from the token.
        """
        del context.block_stack[-1]

        current_end_token = cast(EndMarkdownToken, current_token)
        if not current_end_token.was_forced:
            # We need to do this as the ending fence may be longer than the opening fence.
            assert (
                current_token.extra_data is not None
            ), "extra_data must be defined by this point"
            split_extra_data = current_token.extra_data.split(":")
            assert (
                len(split_extra_data) >= 3
            ), "Extra data must split into at least 3 parts."
            extra_end_space = split_extra_data[1]
            fence_count = int(split_extra_data[2])

            current_start_token = cast(
                FencedCodeBlockMarkdownToken, current_end_token.start_markdown_token
            )

            fence_parts = [
                (
                    ""
                    if previous_token is not None
                    and (
                        previous_token.is_blank_line
                        or previous_token.is_fenced_code_block
                    )
                    else ParserHelper.newline_character
                ),
                current_end_token.extracted_whitespace,
                ParserHelper.repeat_string(
                    current_start_token.fence_character, fence_count
                ),
                extra_end_space,
                ParserHelper.newline_character,
            ]

            return "".join(fence_parts)

        assert (
            previous_token is not None
        ), "Previous token must be defined if processing the end of a fenced block."
        is_previous_code_block = previous_token.is_fenced_code_block
        return (
            ParserHelper.newline_character
            if next_token is not None and not is_previous_code_block
            else ""
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            FencedCodeBlockMarkdownToken,
            FencedCodeBlockMarkdownToken.__handle_start_fenced_code_block_token,
            FencedCodeBlockMarkdownToken.__handle_end_fenced_code_block_token,
        )

    @classmethod
    def __handle_start_fenced_code_block_token(
        cls,
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        start_fence_token = cast(FencedCodeBlockMarkdownToken, next_token)
        token_parts = [output_html]
        if (output_html.endswith("</ol>") or output_html.endswith("</ul>")) or (
            output_html and output_html[-1] != ParserHelper.newline_character
        ):
            token_parts.append(ParserHelper.newline_character)
        transform_state.is_in_code_block, transform_state.is_in_fenced_code_block = (
            True,
            True,
        )
        token_parts.append("<pre><code")
        if start_fence_token.extracted_text:
            token_parts.extend(
                [' class="language-', start_fence_token.extracted_text, '"']
            )
        token_parts.append(">")
        return "".join(token_parts)

    @classmethod
    def __handle_end_fenced_code_block_token(
        cls,
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        fenced_token_index = transform_state.actual_token_index - 1
        while not transform_state.actual_tokens[
            fenced_token_index
        ].is_fenced_code_block:
            fenced_token_index -= 1
        fenced_token = cast(
            FencedCodeBlockMarkdownToken,
            transform_state.actual_tokens[fenced_token_index],
        )

        inner_tag_parts = ["<code"]
        if fenced_token.extracted_text:
            inner_tag_parts.extend(
                [
                    ' class="language-',
                    fenced_token.extracted_text,
                    '"',
                ]
            )
        inner_tag_parts.append(">")
        inner_tag = "".join(inner_tag_parts)

        POGGER.debug(f"inner_tag>>:{inner_tag}:<<")
        POGGER.debug(f"output_html>>:{output_html}:<<")
        POGGER.debug(
            f"last_token>>:{transform_state.actual_tokens[transform_state.actual_token_index - 1]}:<<"
        )

        token_parts = [output_html]
        if (
            not output_html.endswith(inner_tag)
            and output_html[-1] != ParserHelper.newline_character
        ):
            token_parts.append(ParserHelper.newline_character)
            POGGER.debug("#1")
        elif (
            output_html[-1] == ParserHelper.newline_character
            and transform_state.last_token
            and transform_state.last_token.is_text
        ):
            POGGER.debug("#2:$", transform_state.last_token)
            text_token = cast(TextMarkdownToken, transform_state.last_token)
            end_token = cast(EndMarkdownToken, next_token)
            if not (end_token.was_forced and text_token.token_text.endswith("\n\x03")):
                token_parts.append(ParserHelper.newline_character)
        transform_state.is_in_code_block, transform_state.is_in_fenced_code_block = (
            False,
            False,
        )
        token_parts.extend(["</code></pre>", ParserHelper.newline_character])
        return "".join(token_parts)

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if (
            field_name == "fence_character"
            and isinstance(field_value, str)
            and field_value in ["~", "`"]
        ):
            self.__fence_character = field_value
            self.__compose_extra_data_field()
            return True
        return super()._modify_token(field_name, field_value)



"""
Module to provide helper functions for parsing html.
"""

import logging
import string
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.container_blocks.container_helper import ContainerHelper
from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.leaf_blocks.leaf_block_helper import LeafBlockHelper
from pymarkdown.tokens.html_block_markdown_token import HtmlBlockMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import (
    HtmlBlockStackToken,
    ParagraphStackToken,
    StackToken,
)
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class HtmlHelper:
    """
    Class to provide helper functions for parsing html.
    """

    html_block_1 = "1"
    html_block_2 = "2"
    html_block_3 = "3"
    html_block_4 = "4"
    html_block_5 = "5"
    html_block_6 = "6"
    html_block_7 = "7"

    __html_block_start_character = "<"
    __html_tag_start = "/"
    __html_tag_name_end = " >"
    __html_tag_end = ">"
    __html_attribute_value_single = "'"
    __html_attribute_value_double = '"'
    __html_attribute_name_value_separator = "="
    __html_attribute_separator = ParserHelper.space_character
    __valid_tag_name_characters = f"{string.ascii_letters}{string.digits}-"
    __html_block_1_start_tag_prefix = ["script", "pre", "style"]
    __html_tag_attribute_value_terminators = " \"'=<>`"
    __html_block_2_to_5_start = "!"
    __html_block_2_continued_start = "--"
    __html_block_3_continued_start = "?"
    __html_block_4_continued_start = string.ascii_uppercase
    __html_block_5_continued_start = "[CDATA["
    __html_block_1_end_tags = ["</script>", "</pre>", "</style>"]
    __html_block_2_end = "-->"
    __html_block_3_end = "?>"
    __html_block_4_end = __html_tag_end
    __html_block_5_end = "]]>"

    __attribute_start_characters = "abcdefghijklmnopqrstuvwxyz1234567890:_"
    __attribute_other_characters = f"{__attribute_start_characters}.-"

    __html_block_6_start = [
        "address",
        "article",
        "aside",
        "base",
        "basefont",
        "blockquote",
        "body",
        "caption",
        "center",
        "col",
        "colgroup",
        "dd",
        "details",
        "dialog",
        "dir",
        "div",
        "dl",
        "dt",
        "fieldset",
        "figcaption",
        "figure",
        "footer",
        "form",
        "frame",
        "frameset",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "head",
        "header",
        "hr",
        "html",
        "iframe",
        "legend",
        "li",
        "link",
        "main",
        "menu",
        "menuitem",
        "nav",
        "noframes",
        "ol",
        "optgroup",
        "option",
        "p",
        "param",
        "section",
        "source",
        "summary",
        "table",
        "tbody",
        "td",
        "tfoot",
        "th",
        "thead",
        "title",
        "tr",
        "track",
        "ul",
    ]

    @staticmethod
    def is_valid_tag_name(tag_name: str) -> bool:
        """
        Determine if the html tag name is valid according to the html rules.
        """

        return (
            all(
                next_character in HtmlHelper.__valid_tag_name_characters
                for next_character in tag_name.lower()
            )
            if tag_name
            else False
        )

    @staticmethod
    def extract_html_attribute_name(string_to_parse: str, string_index: int) -> int:
        """
        Attempt to extract the attribute name from the provided string.
        """

        string_to_parse_length = len(string_to_parse)
        if not (
            string_index < string_to_parse_length
            and (
                string_to_parse[string_index] in HtmlHelper.__attribute_start_characters
            )
        ):
            return -1

        new_string_index, __ = ParserHelper.collect_while_one_of_characters_verified(
            string_to_parse, string_index + 1, HtmlHelper.__attribute_other_characters
        )
        if new_string_index < string_to_parse_length and string_to_parse[
            new_string_index
        ] in [
            HtmlHelper.__html_attribute_name_value_separator,
            HtmlHelper.__html_attribute_separator,
            HtmlHelper.__html_tag_start,
            HtmlHelper.__html_tag_end,
        ]:
            return new_string_index
        return -1

    @staticmethod
    def extract_optional_attribute_value(line_to_parse: str, value_index: int) -> int:
        """
        Determine and extract an optional attribute value.
        """

        non_whitespace_index, _ = ParserHelper.extract_spaces_verified(
            line_to_parse, value_index
        )
        line_to_parse_size = len(line_to_parse)
        if (
            non_whitespace_index < line_to_parse_size
            and line_to_parse[non_whitespace_index]
            != HtmlHelper.__html_attribute_name_value_separator
        ) or non_whitespace_index >= line_to_parse_size:
            return non_whitespace_index

        non_whitespace_index, _ = ParserHelper.extract_spaces_verified(
            line_to_parse, non_whitespace_index + 1
        )
        if non_whitespace_index < line_to_parse_size:
            first_character_of_value = line_to_parse[non_whitespace_index]
            extracted_text: Optional[str] = None
            if first_character_of_value == HtmlHelper.__html_attribute_value_double:
                (
                    non_whitespace_index,
                    _,
                ) = ParserHelper.collect_until_character_verified(
                    line_to_parse,
                    non_whitespace_index + 1,
                    HtmlHelper.__html_attribute_value_double,
                )
                if non_whitespace_index == line_to_parse_size:
                    return -1
                non_whitespace_index += 1
            elif first_character_of_value == HtmlHelper.__html_attribute_value_single:
                (
                    non_whitespace_index,
                    _,
                ) = ParserHelper.collect_until_character_verified(
                    line_to_parse,
                    non_whitespace_index + 1,
                    HtmlHelper.__html_attribute_value_single,
                )
                if non_whitespace_index == line_to_parse_size:
                    return -1
                non_whitespace_index += 1
            else:
                (
                    non_whitespace_index,
                    extracted_text,
                ) = ParserHelper.collect_until_one_of_characters_verified(
                    line_to_parse,
                    non_whitespace_index,
                    HtmlHelper.__html_tag_attribute_value_terminators,
                )
                if not extracted_text:
                    non_whitespace_index = -1
        else:
            non_whitespace_index = -1
        return non_whitespace_index

    @staticmethod
    def is_complete_html_end_tag(
        tag_name: str, line_to_parse: str, next_char_index: int
    ) -> Tuple[bool, int]:
        """
        Determine if the supplied information is a completed end of tag specification.
        """

        non_whitespace_index, _ = ParserHelper.extract_spaces_verified(
            line_to_parse, next_char_index
        )
        is_valid = HtmlHelper.is_valid_tag_name(tag_name) and (
            non_whitespace_index < len(line_to_parse)
            and line_to_parse[non_whitespace_index] == HtmlHelper.__html_tag_end
        )
        return is_valid, non_whitespace_index + 1

    @staticmethod
    def __is_valid_block_1_tag_name(tag_name: str) -> bool:
        """
        Determine if the tag name is a valid block-1 html tag name.
        """

        return tag_name in HtmlHelper.__html_block_1_start_tag_prefix

    @staticmethod
    def is_complete_html_start_tag(
        tag_name: str, line_to_parse: str, next_char_index: int
    ) -> Tuple[bool, Optional[int]]:
        """
        Determine if the supplied information is a completed start of tag specification.
        """

        is_tag_valid = HtmlHelper.is_valid_tag_name(
            tag_name
        ) and not HtmlHelper.__is_valid_block_1_tag_name(tag_name)

        non_whitespace_index, extracted_whitespace = (
            ParserHelper.extract_spaces_verified(line_to_parse, next_char_index)
        )
        are_attributes_valid: bool = True
        line_to_parse_size: int = len(line_to_parse)
        while (
            is_tag_valid
            and extracted_whitespace
            and are_attributes_valid
            and 0 <= non_whitespace_index < line_to_parse_size
            and line_to_parse[non_whitespace_index]
            not in [HtmlHelper.__html_tag_end, HtmlHelper.__html_tag_start]
        ):
            non_whitespace_index = HtmlHelper.extract_html_attribute_name(
                line_to_parse, non_whitespace_index
            )
            are_attributes_valid = non_whitespace_index != -1
            if not are_attributes_valid:
                break
            non_whitespace_index = HtmlHelper.extract_optional_attribute_value(
                line_to_parse, non_whitespace_index
            )
            are_attributes_valid = non_whitespace_index != -1
            if not are_attributes_valid:
                break
            (
                non_whitespace_index,
                extracted_whitespace,
            ) = ParserHelper.extract_spaces_verified(
                line_to_parse, non_whitespace_index
            )

        if non_whitespace_index < line_to_parse_size:
            if line_to_parse[non_whitespace_index] == HtmlHelper.__html_tag_start:
                non_whitespace_index += 1
            is_end_of_tag_present = (
                line_to_parse[non_whitespace_index] == HtmlHelper.__html_tag_end
            )
            if is_end_of_tag_present:
                non_whitespace_index += 1
        else:
            is_end_of_tag_present = False

        non_whitespace_indexx, _ = ParserHelper.extract_spaces(
            line_to_parse, non_whitespace_index
        )
        return (
            (
                is_tag_valid
                and is_end_of_tag_present
                and non_whitespace_indexx == line_to_parse_size
                and are_attributes_valid
            ),
            non_whitespace_indexx,
        )

    @staticmethod
    def __check_for_special_html_blocks(
        line_to_parse: str, character_index: int
    ) -> Optional[str]:
        """
        Check for the easy to spot special blocks: 2-5.
        """

        if character_index >= len(line_to_parse):
            return None
        html_block_type = None
        if ParserHelper.is_character_at_index(
            line_to_parse, character_index, HtmlHelper.__html_block_2_to_5_start
        ):
            if ParserHelper.are_characters_at_index(
                line_to_parse,
                character_index + 1,
                HtmlHelper.__html_block_2_continued_start,
            ):
                html_block_type = HtmlHelper.html_block_2
            elif ParserHelper.is_character_at_index_one_of(
                line_to_parse,
                character_index + 1,
                HtmlHelper.__html_block_4_continued_start,
            ):
                html_block_type = HtmlHelper.html_block_4
            elif ParserHelper.are_characters_at_index(
                line_to_parse,
                character_index + 1,
                HtmlHelper.__html_block_5_continued_start,
            ):
                html_block_type = HtmlHelper.html_block_5
        elif ParserHelper.is_character_at_index(
            line_to_parse,
            character_index,
            HtmlHelper.__html_block_3_continued_start,
        ):
            html_block_type = HtmlHelper.html_block_3

        return html_block_type

    @staticmethod
    def __check_for_normal_html_blocks_adjust_tag(
        remaining_html_tag: str, line_to_parse: str, character_index: int
    ) -> Tuple[str, int, bool]:
        adjusted_remaining_html_tag, line_to_parse_size = remaining_html_tag, len(
            line_to_parse
        )
        is_end_tag = (
            bool(adjusted_remaining_html_tag)
            and adjusted_remaining_html_tag[0] == HtmlHelper.__html_tag_start
        )
        if is_end_tag:
            adjusted_remaining_html_tag = adjusted_remaining_html_tag[1:]

        if (
            character_index < line_to_parse_size
            and line_to_parse[character_index] == HtmlHelper.__html_tag_end
            and adjusted_remaining_html_tag
            and adjusted_remaining_html_tag[-1] == HtmlHelper.__html_tag_start
        ):
            adjusted_remaining_html_tag = adjusted_remaining_html_tag[:-1]

        return adjusted_remaining_html_tag, line_to_parse_size, is_end_tag

    @staticmethod
    def __check_for_normal_html_blocks(
        remaining_html_tag: str, line_to_parse: str, character_index: int
    ) -> Optional[str]:
        """
        Check for the the html blocks that are harder to identify: 1, 6-7.
        """

        html_block_type = None

        if HtmlHelper.__is_valid_block_1_tag_name(remaining_html_tag):
            html_block_type = HtmlHelper.html_block_1
        else:
            (
                adjusted_remaining_html_tag,
                line_to_parse_size,
                is_end_tag,
            ) = HtmlHelper.__check_for_normal_html_blocks_adjust_tag(
                remaining_html_tag, line_to_parse, character_index
            )

            complete_parse_index: Optional[int] = 0
            if adjusted_remaining_html_tag in HtmlHelper.__html_block_6_start:
                html_block_type = HtmlHelper.html_block_6
            elif is_end_tag:
                is_complete, complete_parse_index = HtmlHelper.is_complete_html_end_tag(
                    adjusted_remaining_html_tag, line_to_parse, character_index
                )
                if is_complete:
                    html_block_type, character_index = (
                        HtmlHelper.html_block_7,
                        complete_parse_index,
                    )
            else:
                (
                    is_complete,
                    complete_parse_index,
                ) = HtmlHelper.is_complete_html_start_tag(
                    adjusted_remaining_html_tag, line_to_parse, character_index
                )
                if is_complete:
                    assert (
                        complete_parse_index is not None
                    ), "If is_complete is True, this must be set."
                    html_block_type, character_index = (
                        HtmlHelper.html_block_7,
                        complete_parse_index,
                    )
            if html_block_type == HtmlHelper.html_block_7:
                new_index, _ = ParserHelper.extract_ascii_whitespace(
                    line_to_parse, character_index
                )
                if new_index != line_to_parse_size:
                    html_block_type = None
        return html_block_type

    @staticmethod
    def __determine_html_block_type(
        token_stack: List[StackToken],
        line_to_parse: str,
        start_index: int,
        parse_properties: ParseBlockPassProperties,
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        Determine the type of the html block that we are starting.
        """

        character_index = start_index + 1
        html_block_type = HtmlHelper.__check_for_special_html_blocks(
            line_to_parse, character_index
        )
        if html_block_type:
            remaining_html_tag = ""
        else:
            (
                new_character_index,
                new_remaining_html_tag,
            ) = ParserHelper.collect_until_one_of_characters_verified(
                line_to_parse, character_index, HtmlHelper.__html_tag_name_end
            )
            remaining_html_tag = new_remaining_html_tag
            character_index = new_character_index
            remaining_html_tag = remaining_html_tag.lower()

            html_block_type = HtmlHelper.__check_for_normal_html_blocks(
                remaining_html_tag, line_to_parse, character_index
            )

        POGGER.debug("html_block_type=$", html_block_type)
        if not html_block_type:
            return None, None
        if html_block_type == HtmlHelper.html_block_7 and token_stack[-1].is_paragraph:
            POGGER.debug("html_block_type 7 cannot interrupt a paragraph")
            return None, None
        if (
            parse_properties.is_disallow_raw_html_enabled
            and remaining_html_tag
            and parse_properties.disallow_raw_html
            and parse_properties.disallow_raw_html.is_html_tag_disallowed(
                remaining_html_tag
            )
        ):
            POGGER.debug("Tag name '%s' is disallowed.", remaining_html_tag)
            return None, None
        return html_block_type, remaining_html_tag

    # pylint: disable=too-many-arguments
    @staticmethod
    def is_html_block(
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        token_stack: List[StackToken],
        parse_properties: ParseBlockPassProperties,
        skip_whitespace_check: bool = False,
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        Determine if the current sequence of characters would start a html block element.
        """

        if (
            skip_whitespace_check
            or TabHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
        ) and ParserHelper.is_character_at_index(
            line_to_parse,
            start_index,
            HtmlHelper.__html_block_start_character,
        ):
            return HtmlHelper.__determine_html_block_type(
                token_stack, line_to_parse, start_index, parse_properties
            )
        return None, None

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_split_tab(
        parser_state: ParserState,
        position_marker: PositionMarker,
        original_line: str,
        orignal_line_end_prefix_index: int,
    ) -> Tuple[Optional[str], Optional[int]]:
        stack_token_index = len(parser_state.token_stack) - 1
        # assert stack_token_index > 0, "Must be a valid stack token."
        # Can omit check as splits only happen with containers, which guarantees one token.

        alternate_list_leading_space = None
        removed_chars_at_start: Optional[int] = None
        # while (
        #     stack_token_index > 0
        #     and not parser_state.token_stack[
        #         stack_token_index
        #     ].is_block_quote
        #     and not parser_state.token_stack[stack_token_index].is_list
        # ):
        #     stack_token_index -= 1
        if parser_state.token_stack[stack_token_index].is_list:
            orignal_line_prefix = original_line[:orignal_line_end_prefix_index]
            stop_index = -1
            if len(orignal_line_prefix) == 1:
                assert (
                    orignal_line_prefix == ParserHelper.tab_character
                ), "Prefix must be a single tab character."
                sdddd = TabHelper.detabify_string(orignal_line_prefix)
                assert (
                    len(sdddd) > position_marker.index_indent
                ), "String length must be larger than the indent."
                stop_index = 1
            else:
                end_index = 1
                keep_going = True
                while keep_going:
                    assert (
                        end_index < len(orignal_line_prefix) + 1
                    ), "End index not in range."
                    sample_slice = TabHelper.detabify_string(
                        orignal_line_prefix[:end_index]
                    )
                    if len(sample_slice) > position_marker.index_indent:
                        stop_index = end_index
                        keep_going = False
                    end_index += 1
                assert stop_index != -1, "Valid slice not found."

            original_prefix = original_line[: stop_index - 1]
            if stack_token_index <= 1:
                alternate_list_leading_space = original_prefix
            removed_chars_at_start = len(original_prefix)
        return alternate_list_leading_space, removed_chars_at_start

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __found_html_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        new_tokens: List[MarkdownToken],
        original_line: str,
        extracted_whitespace: str,
        block_quote_data: BlockQuoteData,
        html_block_type: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[List[MarkdownToken], bool, Optional[int], str]:
        split_tab = False
        alternate_list_leading_space: Optional[str] = None
        removed_chars_at_start: Optional[int] = None
        if ParserHelper.tab_character in original_line:
            token_text = position_marker.text_to_parse[position_marker.index_number :]
            # POGGER.debug("token_text=:$:", token_text)
            (
                _,
                split_tab,
                _,  # split_tab_with_block_quote_suffix,
                _,  # tabified_prefix,
                _,  # tabified_suffix,
                orignal_line_end_prefix_index,
            ) = TabHelper.parse_thematic_break_with_tab(
                original_line, token_text, extracted_whitespace
            )
            # POGGER.debug("split_tab=:$:", split_tab)
            if split_tab:
                (
                    alternate_list_leading_space,
                    removed_chars_at_start,
                ) = HtmlHelper.__handle_split_tab(
                    parser_state,
                    position_marker,
                    original_line,
                    orignal_line_end_prefix_index,
                )

        # POGGER.debug("did_adjust_block_quote=$", did_adjust_block_quote)
        # POGGER.debug("split_tab=$", split_tab)
        old_split_tab = split_tab
        did_adjust_block_quote = False
        split_tab, extracted_whitespace, whitespace_prefix = (
            ContainerHelper.reduce_containers_if_required(
                parser_state,
                position_marker,
                block_quote_data,
                new_tokens,
                split_tab,
                extracted_whitespace,
                grab_bag,
            )
        )
        if split_tab:
            TabHelper.adjust_block_quote_indent_for_tab(
                parser_state,
                extracted_whitespace,
                alternate_list_leading_space,
                original_line=original_line,
            )
            did_adjust_block_quote = True
            POGGER.debug("did_adjust_block_quote=$", did_adjust_block_quote)
        POGGER.debug("split_tab=$", split_tab)
        did_adjust_block_quote = split_tab != old_split_tab or did_adjust_block_quote

        if whitespace_prefix:
            extracted_whitespace = whitespace_prefix + extracted_whitespace

        new_token = HtmlBlockMarkdownToken(position_marker, extracted_whitespace)
        new_tokens.append(new_token)
        parser_state.token_stack.append(HtmlBlockStackToken(html_block_type, new_token))
        return (
            new_tokens,
            did_adjust_block_quote,
            removed_chars_at_start,
            extracted_whitespace,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def parse_html_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        block_quote_data: BlockQuoteData,
        original_line: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[List[MarkdownToken], bool, Optional[int], str]:
        """
        Determine if we have the criteria that we need to start an HTML block.
        """

        check_ws = LeafBlockHelper.realize_leading_whitespace(
            parser_state, position_marker, extracted_whitespace, original_line
        )
        html_block_type, _ = HtmlHelper.is_html_block(
            position_marker.text_to_parse,
            position_marker.index_number,
            check_ws,
            parser_state.token_stack,
            parser_state.parse_properties,
        )
        removed_chars_at_start = None
        did_adjust_block_quote = False
        POGGER.debug("did_adjust_block_quote=$", did_adjust_block_quote)
        if html_block_type:
            new_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[ParagraphStackToken],
            )
            POGGER.debug("new_tokens=$", new_tokens)
            POGGER.debug("original_line=:$:", original_line)
            POGGER.debug("extracted_whitespace=:$:", extracted_whitespace)

            (
                new_tokens,
                did_adjust_block_quote,
                removed_chars_at_start,
                extracted_whitespace,
            ) = HtmlHelper.__found_html_block(
                parser_state,
                position_marker,
                new_tokens,
                original_line,
                extracted_whitespace,
                block_quote_data,
                html_block_type,
                grab_bag,
            )
        else:
            new_tokens = []
        POGGER.debug("did_adjust_block_quote=$", did_adjust_block_quote)
        return (
            new_tokens,
            did_adjust_block_quote,
            removed_chars_at_start,
            extracted_whitespace,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def check_blank_html_block_end(parser_state: ParserState) -> List[MarkdownToken]:
        """
        Check to see if we have encountered the end of the current HTML block
        via an empty line or BLANK.
        """

        assert parser_state.token_stack[
            -1
        ].is_html_block, "Trailing token on stack must be HTML."
        html_token = cast(HtmlBlockStackToken, parser_state.token_stack[-1])
        if html_token.html_block_type in [
            HtmlHelper.html_block_6,
            HtmlHelper.html_block_7,
        ]:
            new_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[HtmlBlockStackToken],
            )
            POGGER.debug("new_tokens=$", new_tokens)
        else:
            new_tokens = []

        return new_tokens

    @staticmethod
    def __check_normal_html_block_end_with_tab(
        parser_state: ParserState,
        original_line: str,
        extracted_whitespace: str,
        token_text: str,
        did_adjust_block_quote: bool,
    ) -> Tuple[str, str]:
        POGGER.debug("did_adjust_block_quote>:$:<", did_adjust_block_quote)
        POGGER.debug("original_line>:$:<", original_line)
        POGGER.debug("token_text>:$:<", token_text)
        POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        (
            tabified_text,
            split_tab,
            split_tab_with_block_quote_suffix,
            _,
            tabified_whitespace,
            _,
        ) = TabHelper.parse_thematic_break_with_tab(
            original_line, token_text, extracted_whitespace
        )

        POGGER.debug("tabified_text>:$:<", tabified_text)
        POGGER.debug("split_tab>:$:<", split_tab)
        POGGER.debug(
            "split_tab_with_block_quote_suffix>:$:<", split_tab_with_block_quote_suffix
        )
        POGGER.debug("tabified_whitespace>:$:<", tabified_whitespace)

        stack_token_index = len(parser_state.token_stack) - 1
        while (
            stack_token_index > 0
            and not parser_state.token_stack[stack_token_index].is_block_quote
            and not parser_state.token_stack[stack_token_index].is_list
        ):
            stack_token_index -= 1

        if (
            split_tab
            and stack_token_index != 0
            and parser_state.token_stack[stack_token_index].is_block_quote
        ):
            POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
            tabified_whitespace = ParserHelper.create_replacement_markers(
                tabified_whitespace, extracted_whitespace
            )
            POGGER.debug("tabified_whitespace>:$:<", tabified_whitespace)
            if not did_adjust_block_quote:
                TabHelper.adjust_block_quote_indent_for_tab(parser_state)
        return tabified_whitespace, tabified_text

    @staticmethod
    def __handle_disallow(parser_state: ParserState, token_text: str) -> str:
        cleaned_up_text_parts: List[str] = []
        new_index, new_index_text = ParserHelper.collect_until_character_verified(
            token_text, 0, HtmlHelper.__html_block_start_character
        )
        x_index: int = new_index
        while x_index < len(token_text):
            cleaned_up_text_parts.append(new_index_text)
            (
                after_text_index,
                collected_text,
            ) = ParserHelper.collect_until_one_of_characters_verified(
                token_text, new_index + 1, " /<>"
            )
            if (
                after_text_index + 1 < len(token_text)
                and token_text[after_text_index] == "<"
            ):
                cleaned_up_text_parts.append(f"<{collected_text}")
                new_index, new_index_text = (
                    ParserHelper.collect_until_character_verified(
                        token_text,
                        after_text_index,
                        HtmlHelper.__html_block_start_character,
                    )
                )
                continue
            assert (
                parser_state.parse_properties is not None
                and parser_state.parse_properties.disallow_raw_html is not None
            ), "Disallow raw html extension must be defined by this point."
            tag_start = (
                ParserHelper.create_replacement_markers(
                    HtmlHelper.__html_block_start_character, "&lt;"
                )
                if parser_state.parse_properties.disallow_raw_html.is_html_tag_disallowed(
                    collected_text
                )
                else HtmlHelper.__html_block_start_character
            )
            cleaned_up_text_parts.append(tag_start + collected_text)
            new_index, new_index_text = ParserHelper.collect_until_character_verified(
                token_text,
                after_text_index,
                HtmlHelper.__html_block_start_character,
            )
            x_index = new_index
        cleaned_up_text_parts.append(new_index_text)
        return "".join(cleaned_up_text_parts)

    # pylint: disable=too-many-arguments
    @staticmethod
    def check_normal_html_block_end(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        position_marker: PositionMarker,
        original_line: str,
        did_adjust_block_quote: bool,
    ) -> List[MarkdownToken]:
        """
        Check to see if we have encountered the end of the current HTML block
        via text on a normal line.
        """

        token_text = line_to_parse[start_index:]
        # POGGER.debug("extracted_whitespace=:$:", extracted_whitespace)
        # POGGER.debug("token_text=:$:", token_text)
        # POGGER.debug("original_line=:$:", original_line)
        if ParserHelper.tab_character in original_line:
            (
                extracted_whitespace,
                token_text,
            ) = HtmlHelper.__check_normal_html_block_end_with_tab(
                parser_state,
                original_line,
                extracted_whitespace,
                token_text,
                did_adjust_block_quote,
            )

        if parser_state.parse_properties.is_disallow_raw_html_enabled:
            token_text = HtmlHelper.__handle_disallow(parser_state, token_text)

        new_tokens: List[MarkdownToken] = [
            TextMarkdownToken(
                token_text,
                extracted_whitespace,
                position_marker=position_marker,
            )
        ]

        is_block_terminated, adj_line = False, line_to_parse[start_index:]
        assert parser_state.token_stack[
            -1
        ].is_html_block, "Trailing token must be HTML token."
        html_token = cast(HtmlBlockStackToken, parser_state.token_stack[-1])
        if html_token.html_block_type == HtmlHelper.html_block_1:
            for next_end_tag in HtmlHelper.__html_block_1_end_tags:
                if next_end_tag in adj_line:
                    is_block_terminated = True
        elif html_token.html_block_type == HtmlHelper.html_block_2:
            is_block_terminated = HtmlHelper.__html_block_2_end in adj_line
        elif html_token.html_block_type == HtmlHelper.html_block_3:
            is_block_terminated = HtmlHelper.__html_block_3_end in adj_line
        elif html_token.html_block_type == HtmlHelper.html_block_4:
            is_block_terminated = HtmlHelper.__html_block_4_end in adj_line
        elif html_token.html_block_type == HtmlHelper.html_block_5:
            is_block_terminated = HtmlHelper.__html_block_5_end in adj_line

        if is_block_terminated:
            terminated_block_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[HtmlBlockStackToken],
            )
            # POGGER.debug("terminated_block_tokens=$", terminated_block_tokens)
            assert terminated_block_tokens, "At least one token must be produced."
            new_tokens.extend(terminated_block_tokens)
        return new_tokens

    # pylint: enable=too-many-arguments



"""
Module to provide for an encapsulation of the html block element.
"""

import logging
from typing import Optional

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class HtmlBlockMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the html block element.
    """

    def __init__(
        self, position_marker: PositionMarker, extracted_whitespace: str
    ) -> None:
        line_number = position_marker.line_number if position_marker else -1
        column_number = (
            position_marker.index_number
            + position_marker.index_indent
            + 1
            - len(extracted_whitespace)
            if position_marker
            else -1
        )
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_html_block,
            "",
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_html_block

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            HtmlBlockMarkdownToken,
            HtmlBlockMarkdownToken.__rehydrate_html_block,
            HtmlBlockMarkdownToken.__rehydrate_html_block_end,
        )

    @staticmethod
    def __rehydrate_html_block(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the html block from the token.
        """
        _ = (current_token, previous_token)

        context.block_stack.append(current_token)
        return ""

    @staticmethod
    def __rehydrate_html_block_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the end of the html block from the token.
        """
        _ = (current_token, previous_token, next_token)

        del context.block_stack[-1]
        return ""

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            HtmlBlockMarkdownToken,
            HtmlBlockMarkdownToken.__handle_start_html_block_token,
            HtmlBlockMarkdownToken.__handle_end_html_block_token,
        )

    @staticmethod
    def __handle_start_html_block_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        transform_state.is_in_html_block = True
        token_parts = []
        if (
            not output_html
            and transform_state.transform_stack
            and transform_state.transform_stack[-1].endswith("<li>")
        ):
            token_parts.append(ParserHelper.newline_character)
        else:
            previous_token = transform_state.actual_tokens[
                transform_state.actual_token_index - 1
            ]
            POGGER.debug(">previous_token>$>", previous_token)
            token_parts.append(output_html)
            if (
                not previous_token.is_list_end
                and previous_token.is_paragraph_end
                and not transform_state.is_in_loose_list
                or previous_token.is_list_end
            ):
                token_parts.append(ParserHelper.newline_character)
        return "".join(token_parts)

    @staticmethod
    def __handle_end_html_block_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        transform_state.is_in_html_block = False
        return output_html



from typing import Dict, List, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.reference_markdown_token import ReferenceMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class ListTracker:
    """
    Class to help rules keep track of lists.
    """

    def __init__(self) -> None:
        self.__list_stack: List[ListStartMarkdownToken] = []
        self.__line_count: Dict[int, int] = {}
        self.__list_start_indices: Dict[int, Dict[MarkdownToken, int]] = {}
        self.__list_end_indices: Dict[int, Dict[MarkdownToken, int]] = {}
        self.__list_adjustments: Dict[int, Dict[MarkdownToken, int]] = {}
        self.__current_list_tokens: Dict[int, MarkdownToken] = {}

        self.__was_last_blank_line: bool = False

    def starting_new_file(self) -> None:
        """
        Called from the `starting_new_file` function of the rule.
        """
        self.__line_count = {}
        self.__list_stack = []
        self.__list_start_indices = {}
        self.__list_end_indices = {}
        self.__list_adjustments = {}
        self.__current_list_tokens = {}
        self.__was_last_blank_line = False

    def next_token(self, token: MarkdownToken) -> None:
        """
        Called from the `next_token` function of the rule.
        """

        # Weird calculation.  A blank line terminates the previous line and is
        # blank line on its own... EXCEPT if it is not followed by an inline.
        list_stack_length = len(self.__list_stack)
        x = True
        if self.__was_last_blank_line and not token.is_end_token:
            if not token.is_container and list_stack_length:
                self.__line_count[list_stack_length] += 1
                x = False
            self.__was_last_blank_line = False
        if token.is_blank_line:
            self.__was_last_blank_line = True

        if list_stack_length and x:
            self.__line_count[list_stack_length] += self.__count_newlines_in_token(
                token
            )
        x = False

    def register(self, token: MarkdownToken, register_value: int) -> None:
        """
        Register an integer for a given list token at the current level.
        """
        list_level = len(self.__list_stack)
        self.__list_adjustments[list_level][token] = register_value

    def get_registrations(self) -> Dict[MarkdownToken, int]:
        """
        Get any registrations at this level.
        """
        list_level = len(self.__list_stack)
        return self.__list_adjustments[list_level]

    def get_start_stop(self, token: MarkdownToken) -> Tuple[int, int]:
        """
        Get the start and stop positions of the current list item.
        """
        list_level = len(self.__list_stack)
        start_index = self.__list_start_indices[list_level][token]
        end_index = self.__list_end_indices[list_level][token]
        return start_index, end_index

    def list_start(self, token: MarkdownToken) -> None:
        """
        Record the start of a list.
        """
        assert isinstance(token, ListStartMarkdownToken)
        self.__list_stack.append(token)
        list_level = len(self.__list_stack)
        self.__line_count[list_level] = 0

        new_map: Dict[MarkdownToken, int] = {}
        self.__list_start_indices[list_level] = new_map
        new_map[token] = self.__line_count[list_level]

        new_map = {}
        self.__list_end_indices[list_level] = new_map
        self.__current_list_tokens[list_level] = token

        new_map = {}
        self.__list_adjustments[list_level] = new_map

    def new_list_item(self, token: MarkdownToken) -> None:
        """
        Record a new item within the list.
        """
        list_level = len(self.__list_stack)
        assert list_level in self.__list_start_indices
        current_list_token = self.__current_list_tokens[list_level]
        self.__current_list_tokens[list_level] = token

        self.__list_end_indices[list_level][current_list_token] = self.__line_count[
            list_level
        ]
        self.__list_start_indices[list_level][token] = self.__line_count[list_level]
        self.__was_last_blank_line = False

    def list_end(self) -> None:
        """
        Record the end of a list.
        """
        list_level = len(self.__list_stack)
        assert list_level in self.__list_start_indices

        current_list_token = self.__current_list_tokens[list_level]
        self.__list_end_indices[list_level][current_list_token] = self.__line_count[
            list_level
        ]

    def list_end_cleanup(self) -> None:
        """
        Cleanup after any processing of the end of the list.
        """
        list_level = len(self.__list_stack)
        assert list_level in self.__list_start_indices

        del self.__list_start_indices[list_level]
        del self.__list_end_indices[list_level]
        del self.__current_list_tokens[list_level]
        del self.__list_adjustments[list_level]

        del self.__line_count[len(self.__list_stack)]
        del self.__list_stack[-1]

    def __count_newlines_in_token(self, current_token: MarkdownToken) -> int:
        newlines_in_text_token = 0
        if current_token.is_blank_line:
            newlines_in_text_token = 1
        elif current_token.is_text:
            text_token = cast(TextMarkdownToken, current_token)
            newlines_in_text_token = ParserHelper.count_newlines_in_text(
                text_token.token_text
            )
        elif current_token.is_inline_image or current_token.is_inline_link:
            reference_token = cast(ReferenceMarkdownToken, current_token)
            newlines_in_text_token += ParserHelper.count_newlines_in_text(
                reference_token.text_from_blocks
            )
        elif current_token.is_inline_raw_html:
            raw_html_token = cast(RawHtmlMarkdownToken, current_token)
            newlines_in_text_token += ParserHelper.count_newlines_in_text(
                raw_html_token.raw_tag
            )
        elif current_token.is_inline_code_span:
            code_span_token = cast(InlineCodeSpanMarkdownToken, current_token)
            newlines_in_text_token += ParserHelper.count_newlines_in_texts(
                code_span_token.leading_whitespace,
                code_span_token.span_text,
                code_span_token.trailing_whitespace,
            )
        return newlines_in_text_token



"""
Module to implement a plugin that looks for heading styles that are inconsistent
throughout the document.
"""

from typing import List, Tuple, cast

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd003(RulePlugin):
    """
    Class to implement a plugin that looks for heading styles that are inconsistent
    throughout the document.
    """

    __consistent_style = "consistent"
    __atx_style = "atx"
    __atx_closed_style = "atx_closed"
    __setext_style = "setext"
    __setext_with_atx_style = "setext_with_atx"
    __setext_with_atx_closed_style = "setext_with_atx_closed"

    __simple_styles = [__atx_style, __atx_closed_style, __setext_style]
    __valid_styles = [
        __consistent_style,
        __atx_style,
        __atx_closed_style,
        __setext_style,
        __setext_with_atx_style,
        __setext_with_atx_closed_style,
    ]

    def __init__(self) -> None:
        super().__init__()
        self.__style_type: str = ""
        self.__actual_style_type: str = ""
        self.__allow_consistent_setext_update = False

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="heading-style,header-style",
            plugin_id="MD003",
            plugin_enabled_by_default=True,
            plugin_description="Heading style should be consistent throughout the document.",
            plugin_version="0.6.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md003.md",
            plugin_configuration="style",
        )

    @classmethod
    def __validate_configuration_style(cls, found_value: str) -> None:
        if found_value not in RuleMd003.__valid_styles:
            raise ValueError(f"Allowable values: {RuleMd003.__valid_styles}")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__style_type = self.plugin_configuration.get_string_property_with_default(
            "style",
            RuleMd003.__consistent_style,
            valid_value_fn=self.__validate_configuration_style,
        )
        self.__allow_consistent_setext_update = (
            self.plugin_configuration.get_boolean_property_with_default(
                "allow-setext-update", False
            )
            if self.__style_type == RuleMd003.__consistent_style
            else False
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("style", self.__style_type),
            QueryConfigItem(
                "allow-setext-update", self.__allow_consistent_setext_update
            ),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__actual_style_type = (
            self.__style_type
            if self.__style_type != RuleMd003.__consistent_style
            else ""
        )

    def __handle_simple_styles(
        self, heading_style_type: str, is_heading_level_1_or_2: bool
    ) -> Tuple[bool, str]:
        is_heading_bad = heading_style_type != self.__actual_style_type

        if (
            is_heading_bad
            and self.__allow_consistent_setext_update
            and self.__actual_style_type == RuleMd003.__setext_style
            and heading_style_type == RuleMd003.__atx_style
            and not is_heading_level_1_or_2
        ):
            is_heading_bad = False
            self.__actual_style_type = RuleMd003.__setext_with_atx_style

        expected_style_type = self.__actual_style_type
        return is_heading_bad, expected_style_type

    def __handle_complex_styles(
        self, heading_style_type: str, is_heading_level_1_or_2: bool
    ) -> Tuple[bool, str]:
        is_heading_bad, expected_style_type = False, ""
        if self.__actual_style_type == RuleMd003.__setext_with_atx_style:
            base_atx_style = RuleMd003.__atx_style
        else:
            assert self.__actual_style_type == RuleMd003.__setext_with_atx_closed_style
            base_atx_style = RuleMd003.__atx_closed_style
        if not (
            (is_heading_level_1_or_2 and heading_style_type == RuleMd003.__setext_style)
            or (not is_heading_level_1_or_2 and heading_style_type == base_atx_style)
        ):
            is_heading_bad = True
            expected_style_type = (
                RuleMd003.__setext_style if is_heading_level_1_or_2 else base_atx_style
            )
            if (
                expected_style_type == RuleMd003.__setext_style
                and is_heading_level_1_or_2
                and self.__actual_style_type == RuleMd003.__setext_with_atx_style
            ):
                expected_style_type = RuleMd003.__setext_with_atx_style
        return is_heading_bad, expected_style_type

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        heading_style_type, is_heading_level_1_or_2 = self.__get_heading_properties(
            token
        )

        is_heading_bad = False
        if heading_style_type:
            expected_style_type = None
            if not self.__actual_style_type:
                self.__actual_style_type = heading_style_type
            elif self.__actual_style_type in RuleMd003.__simple_styles:
                is_heading_bad, expected_style_type = self.__handle_simple_styles(
                    heading_style_type, is_heading_level_1_or_2
                )
            else:
                is_heading_bad, expected_style_type = self.__handle_complex_styles(
                    heading_style_type, is_heading_level_1_or_2
                )

            if is_heading_bad:
                extra_data = (
                    f"Expected: {expected_style_type}; Actual: {heading_style_type}"
                )
                self.report_next_token_error(
                    context,
                    token,
                    extra_error_information=extra_data,
                    use_original_position=token.is_setext_heading,
                )

    @classmethod
    def __get_heading_properties(cls, token: MarkdownToken) -> Tuple[str, bool]:
        """
        Determine the heading properties related to the current token.
        """

        heading_style_type = ""
        is_heading_level_1_or_2 = False
        if token.is_atx_heading:
            atx_token = cast(AtxHeadingMarkdownToken, token)
            heading_style_type = (
                RuleMd003.__atx_closed_style
                if atx_token.remove_trailing_count
                else RuleMd003.__atx_style
            )
            is_heading_level_1_or_2 = atx_token.hash_count < 3
        elif token.is_setext_heading:
            heading_style_type = RuleMd003.__setext_style
            is_heading_level_1_or_2 = True
        return heading_style_type, is_heading_level_1_or_2



"""
Module to helper with the parsing of link reference definitions.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.links.link_reference_definition_continuation_helper import (
    LinkReferenceDefinitionContinuationHelper,
)
from pymarkdown.links.link_reference_definition_parse_helper import (
    LinkReferenceDefinitionParseHelper,
)
from pymarkdown.links.link_reference_tuple import LinkReferenceDefinitionTuple
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import LinkDefinitionStackToken

POGGER = ParserLogger(logging.getLogger(__name__))


class LinkReferenceDefinitionHelper:
    """
    Class to helper with the parsing of link reference definitions.
    """

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def process_link_reference_definition(
        parser_state: ParserState,
        position_marker: PositionMarker,
        remaining_line_to_parse: str,
        extracted_whitespace: str,
        unmodified_line_to_parse: str,
        original_stack_depth: int,
        original_document_depth: int,
        original_line: str,
    ) -> Tuple[bool, bool, bool, Optional[RequeueLineInfo], List[MarkdownToken]]:
        """
        Process a link deference definition.  Note, this requires a lot of work to
        handle properly because of partial definitions across lines.
        """
        line_to_parse = position_marker.text_to_parse
        start_index: int = position_marker.index_number
        lines_to_requeue: List[str] = []

        POGGER.debug(">>remaining_line_to_parse>:$:<", remaining_line_to_parse)
        POGGER.debug(">>line_to_parse>:$:<", line_to_parse)
        POGGER.debug(">>start_index>:$:<", start_index)

        (
            line_to_parse,
            unmodified_line_to_parse,
            remaining_line_to_parse,
            start_index,
            is_blank_line,
        ) = LinkReferenceDefinitionHelper.__handle_link_reference_definition_init(
            remaining_line_to_parse,
            line_to_parse,
            start_index,
            original_line,
            unmodified_line_to_parse,
        )

        (
            was_started,
            lrd_stack_token,
            original_stack_depth,
            original_document_depth,
            line_to_parse,
            extracted_whitespace,
            start_index,
        ) = LinkReferenceDefinitionHelper.__handle_link_reference_definition_started(
            parser_state,
            original_stack_depth,
            original_document_depth,
            line_to_parse,
            extracted_whitespace,
            start_index,
        )
        POGGER.debug(">>line_to_parse>:$:<", line_to_parse)
        line_to_parse_size = len(line_to_parse)

        (
            did_complete_lrd,
            end_lrd_index,
            parsed_lrd_tuple,
            is_blank_line,
            line_to_parse,
        ) = LinkReferenceDefinitionHelper.__handle_link_reference_definition_processing(
            parser_state,
            line_to_parse,
            start_index,
            extracted_whitespace,
            is_blank_line,
            was_started,
            line_to_parse_size,
            remaining_line_to_parse,
            lines_to_requeue,
            unmodified_line_to_parse,
        )
        (
            did_pause_lrd,
            force_ignore_first_as_lrd,
            new_tokens,
        ) = LinkReferenceDefinitionContinuationHelper.determine_continue_or_stop(
            parser_state,
            position_marker,
            was_started,
            remaining_line_to_parse,
            extracted_whitespace,
            unmodified_line_to_parse,
            original_stack_depth,
            original_document_depth,
            end_lrd_index,
            line_to_parse_size,
            is_blank_line,
            did_complete_lrd,
            parsed_lrd_tuple,
            lines_to_requeue,
        )
        if lines_to_requeue:
            assert (
                lrd_stack_token is not None
            ), "Stack token is created before determining requeue, therefore it must be declared."
            LinkReferenceDefinitionHelper.__prepare_for_requeue(
                parser_state,
                lrd_stack_token,
                did_complete_lrd,
                original_stack_depth,
                original_document_depth,
                lines_to_requeue,
            )
            requeue_line_info = RequeueLineInfo(
                lines_to_requeue, force_ignore_first_as_lrd, False
            )
        else:
            requeue_line_info = None

        return (
            did_complete_lrd or end_lrd_index != -1,
            did_complete_lrd,
            did_pause_lrd,
            requeue_line_info,
            new_tokens,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __handle_link_reference_definition_init(
        remaining_line_to_parse: str,
        line_to_parse: str,
        start_index: int,
        original_line: str,
        unmodified_line_to_parse: str,
    ) -> Tuple[str, str, str, int, bool]:
        is_blank_line = not line_to_parse and not start_index

        POGGER.debug(">>original_line>:$:<", original_line)
        if ParserHelper.tab_character in original_line and not is_blank_line:
            POGGER.debug(">>tabified>:$:<", original_line)

            first_character_to_parse = line_to_parse[start_index]
            POGGER.debug(">>xx>:$:<", first_character_to_parse)
            first_character_to_parse_index = original_line.find(
                first_character_to_parse
            )
            assert (
                first_character_to_parse_index != -1
            ), "Character not found within string."

            line_to_parse = original_line[first_character_to_parse_index:]
            unmodified_line_to_parse = original_line
            remaining_line_to_parse = line_to_parse
            start_index = 0

        POGGER.debug(">>line_to_parse>:$:<", line_to_parse)
        return (
            line_to_parse,
            unmodified_line_to_parse,
            remaining_line_to_parse,
            start_index,
            is_blank_line,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_link_reference_definition_started(
        parser_state: ParserState,
        original_stack_depth: int,
        original_document_depth: int,
        line_to_parse: str,
        extracted_whitespace: str,
        start_index: int,
    ) -> Tuple[bool, Optional[LinkDefinitionStackToken], int, int, str, str, int]:
        lrd_stack_token: Optional[LinkDefinitionStackToken] = None

        if was_started := parser_state.token_stack[-1].was_link_definition_started:
            lrd_stack_token = cast(
                LinkDefinitionStackToken, parser_state.token_stack[-1]
            )
            assert (
                lrd_stack_token.original_stack_depth is not None
                and lrd_stack_token.original_document_depth is not None
            ), "stack and document depth must both be defined."
            original_stack_depth, original_document_depth = (
                lrd_stack_token.original_stack_depth,
                lrd_stack_token.original_document_depth,
            )
            POGGER.debug(
                ">>continuation_lines>>$<<",
                lrd_stack_token.continuation_lines,
            )
            line_to_parse = lrd_stack_token.add_joined_lines_before_suffix(
                line_to_parse
            )
            (
                start_index,
                extracted_whitespace,
            ) = ParserHelper.extract_ascii_whitespace_verified(line_to_parse, 0)

            POGGER.debug(">>line_to_parse>>$<<", line_to_parse)
        return (
            was_started,
            lrd_stack_token,
            original_stack_depth,
            original_document_depth,
            line_to_parse,
            extracted_whitespace,
            start_index,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_link_reference_definition_processing(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        is_blank_line: bool,
        was_started: bool,
        line_to_parse_size: int,
        remaining_line_to_parse: str,
        lines_to_requeue: List[str],
        unmodified_line_to_parse: str,
    ) -> Tuple[bool, int, Optional[LinkReferenceDefinitionTuple], bool, str]:
        if was_started:
            POGGER.debug(">>parse_link_reference_definition>>was_started")

            (
                did_complete_lrd,
                end_lrd_index,
                parsed_lrd_tuple,
            ) = LinkReferenceDefinitionParseHelper.parse_link_reference_definition(
                parser_state,
                line_to_parse,
                start_index,
                extracted_whitespace,
                is_blank_line,
                remaining_line_to_parse,
                was_started,
            )
            POGGER.debug(
                ">>parse_link_reference_definition>>was_started>>did_complete_lrd>>$"
                + ">>end_lrd_index>>$>>len(line_to_parse)>>$",
                did_complete_lrd,
                end_lrd_index,
                line_to_parse_size,
            )

            if not (
                did_complete_lrd
                or (
                    not is_blank_line
                    and not did_complete_lrd
                    and (end_lrd_index == line_to_parse_size)
                )
            ):
                POGGER.debug(
                    ">>parse_link_reference_definition>>was_started>>GOT HARD FAILURE"
                )
                (
                    is_blank_line,
                    line_to_parse,
                    did_complete_lrd,
                    end_lrd_index,
                    parsed_lrd_tuple,
                ) = LinkReferenceDefinitionHelper.__process_lrd_hard_failure(
                    parser_state,
                    remaining_line_to_parse,
                    lines_to_requeue,
                    unmodified_line_to_parse,
                    was_started,
                )
        else:
            (
                did_complete_lrd,
                end_lrd_index,
                parsed_lrd_tuple,
            ) = LinkReferenceDefinitionParseHelper.parse_link_reference_definition(
                parser_state,
                line_to_parse,
                start_index,
                extracted_whitespace,
                is_blank_line,
                remaining_line_to_parse,
                was_started,
            )
            POGGER.debug(
                ">>parse_link_reference_definition>>did_complete_lrd>>$>>end_lrd_index>>$>>len(line_to_parse)>>$",
                did_complete_lrd,
                end_lrd_index,
                line_to_parse_size,
            )
        return (
            did_complete_lrd,
            end_lrd_index,
            parsed_lrd_tuple,
            is_blank_line,
            line_to_parse,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __prepare_for_requeue_reset_markdown_token(
        parser_state: ParserState, lrd_stack_token: LinkDefinitionStackToken
    ) -> None:
        assert (
            lrd_stack_token.last_block_quote_markdown_token_index is not None
        ), "Index must be defined by now."
        POGGER.debug(
            ">>XXXXXX>>last_block_quote_markdown_token_index:$:",
            lrd_stack_token.last_block_quote_markdown_token_index,
        )
        POGGER.debug(
            ">>XXXXXX>>st-now:$:",
            parser_state.token_document[
                lrd_stack_token.last_block_quote_markdown_token_index
            ],
        )

        del parser_state.token_document[
            lrd_stack_token.last_block_quote_markdown_token_index
        ]
        assert (
            lrd_stack_token.copy_of_last_block_quote_markdown_token is not None
        ), "Token must be defined by now."
        parser_state.token_document.insert(
            lrd_stack_token.last_block_quote_markdown_token_index,
            lrd_stack_token.copy_of_last_block_quote_markdown_token,
        )
        assert (
            lrd_stack_token.last_block_quote_stack_token is not None
        ), "Token must be defined by now."
        POGGER.debug(
            "last_block_quote_stack_token>:$:<",
            lrd_stack_token.last_block_quote_stack_token,
        )
        lrd_stack_token.last_block_quote_stack_token.reset_matching_markdown_token(
            lrd_stack_token.copy_of_last_block_quote_markdown_token
        )
        POGGER.debug(
            "last_block_quote_stack_token>:$:<",
            lrd_stack_token.last_block_quote_stack_token,
        )

    @staticmethod
    def __prepare_for_requeue_reset_document_and_stack(
        parser_state: ParserState,
        lrd_stack_token: LinkDefinitionStackToken,
        original_stack_depth: int,
        original_document_depth: int,
    ) -> None:
        POGGER.debug(">>XXXXXX>>original_stack_depth:$:", original_stack_depth)
        POGGER.debug(">>XXXXXX>>token_stack_depth:$:", len(parser_state.token_stack))
        POGGER.debug(">>XXXXXX>>token_stack(before):$:", parser_state.token_stack)
        POGGER.debug(
            ">>XXXXXX>>copy_of_token_stack:$:", parser_state.copy_of_token_stack
        )
        POGGER.debug(
            ">>lrd_stack_token>>copy_of_token_stack:$:",
            lrd_stack_token.copy_of_token_stack,
        )
        if len(parser_state.token_stack) >= original_stack_depth:
            while len(parser_state.token_stack) > original_stack_depth:
                del parser_state.token_stack[-1]
        else:
            while len(parser_state.token_stack):
                del parser_state.token_stack[-1]
            assert (
                lrd_stack_token.copy_of_token_stack is not None
            ), "Token must be defined by now."
            parser_state.token_stack.extend(lrd_stack_token.copy_of_token_stack)
        POGGER.debug(">>XXXXXX>>token_stack(after):$:", parser_state.token_stack)

        LinkReferenceDefinitionHelper.__prepare_for_requeue_reset_document_and_stack_inner(
            parser_state, original_document_depth
        )
        POGGER.debug(">>XXXXXX>>token_document(after):$:", parser_state.token_document)

    @staticmethod
    def __prepare_for_requeue_reset_document_and_stack_inner(
        parser_state: ParserState, original_document_depth: int
    ) -> None:
        POGGER.debug(">>XXXXXX>>original_document_depth:$:", original_document_depth)
        POGGER.debug(
            ">>XXXXXX>>token_document_depth:$:",
            len(parser_state.token_document),
        )
        POGGER.debug(">>XXXXXX>>token_document(before):$:", parser_state.token_document)
        have_had_undeleted_block_stack_token = False
        while len(parser_state.token_document) > original_document_depth:
            have_had_recent_undeleted_block_stack_token = False
            matching_markdown_token = parser_state.token_stack[
                -1
            ].matching_markdown_token
            if (
                matching_markdown_token
                and parser_state.token_document[-1] == matching_markdown_token
                and (
                    matching_markdown_token.is_block_quote_start
                    or matching_markdown_token.is_list_start
                )
            ):
                del parser_state.token_stack[-1]
                have_had_recent_undeleted_block_stack_token = (
                    have_had_undeleted_block_stack_token
                ) = True

            if (
                have_had_undeleted_block_stack_token
                and not have_had_recent_undeleted_block_stack_token
            ):
                assert (
                    len(parser_state.token_document) - 1 == original_document_depth
                    or parser_state.token_document[-1].is_table_end
                )
                break
            del parser_state.token_document[-1]

    # pylint: disable=too-many-arguments
    @staticmethod
    def __prepare_for_requeue(
        parser_state: ParserState,
        lrd_stack_token: LinkDefinitionStackToken,
        did_complete_lrd: bool,
        original_stack_depth: int,
        original_document_depth: int,
        lines_to_requeue: List[str],
    ) -> None:
        # This works because in most cases, we add things.  However, in cases like
        # an indented code block, we process the "is it indented enough" and close
        # that block before hitting this.  As such, we have a special case to take
        # care of that.  In the future, will possibly want to do something instead of
        # original_document_depth and stack, such as passing in a copy of the both
        # elements so they can be reset on the rewind.
        # i.e. icode would go back on stack, end-icode would not be in document.
        POGGER.debug("lines_to_requeue:$:", lines_to_requeue)
        POGGER.debug(
            ">>XXXXXX>>copy_of_last_block_quote_markdown_token:$:",
            lrd_stack_token.copy_of_last_block_quote_markdown_token,
        )
        if not did_complete_lrd:
            if lrd_stack_token.copy_of_last_block_quote_markdown_token:
                LinkReferenceDefinitionHelper.__prepare_for_requeue_reset_markdown_token(
                    parser_state, lrd_stack_token
                )
            LinkReferenceDefinitionHelper.__prepare_for_requeue_reset_document_and_stack(
                parser_state,
                lrd_stack_token,
                original_stack_depth,
                original_document_depth,
            )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __process_lrd_hard_failure(
        parser_state: ParserState,
        remaining_line_to_parse: str,
        lines_to_requeue: List[str],
        unmodified_line_to_parse: str,
        was_started: bool,
    ) -> Tuple[
        bool,
        str,
        bool,
        int,
        Optional[LinkReferenceDefinitionTuple],
    ]:
        """
        In cases of a hard failure, we have had continuations to the original line
        that make it a bit more difficult to figure out if we have an actual good
        LRD in the mix somehow.  So take lines off the end while we have lines.
        """
        (
            is_blank_line,
            line_to_parse,
            did_complete_lrd,
            end_lrd_index,
            parsed_lrd_tuple,
        ) = (None, None, None, None, None)
        link_ref_stack_token = cast(
            LinkDefinitionStackToken, parser_state.token_stack[-1]
        )

        POGGER.debug(">>remaining_line_to_parse>>add>:$<<", remaining_line_to_parse)
        POGGER.debug(">>unmodified_line_to_parse>>add>:$<<", unmodified_line_to_parse)
        assert unmodified_line_to_parse.endswith(
            remaining_line_to_parse
        ), "Current line must end with the remaining text."
        assert unmodified_line_to_parse is not None
        link_ref_stack_token.add_continuation_line(
            parser_state.original_line_to_parse or ""
        )
        if unmodified_line_to_parse or (parser_state.original_line_to_parse is None):
            link_ref_stack_token.add_unmodified_line(unmodified_line_to_parse)
        else:
            link_ref_stack_token.add_unmodified_line(
                parser_state.original_line_to_parse
            )
        while link_ref_stack_token.continuation_lines:
            POGGER.debug(
                "continuation_lines>>$<<",
                ParserHelper.make_whitespace_visible(
                    str(link_ref_stack_token.continuation_lines)
                ),
            )
            POGGER.debug(
                "unmodified_lines>>$<<",
                ParserHelper.make_whitespace_visible(
                    str(link_ref_stack_token.unmodified_lines)
                ),
            )

            assert link_ref_stack_token.unmodified_lines[-1] is not None
            lines_to_requeue.append(link_ref_stack_token.unmodified_lines[-1])
            POGGER.debug(
                ">>continuation_line>>$",
                link_ref_stack_token.continuation_lines[-1],
            )
            POGGER.debug(
                ">>unmodified_line>>$",
                link_ref_stack_token.unmodified_lines[-1],
            )
            del link_ref_stack_token.continuation_lines[-1]
            del link_ref_stack_token.unmodified_lines[-1]
            POGGER.debug(
                ">>lines_to_requeue>>$>>",
                lines_to_requeue,
            )
            POGGER.debug(
                ">>continuation_lines>>$<<",
                link_ref_stack_token.continuation_lines,
            )
            POGGER.debug(
                ">>unmodified_lines>>$<<",
                link_ref_stack_token.unmodified_lines,
            )
            (
                is_blank_line,
                line_to_parse,
            ) = True, link_ref_stack_token.add_joined_lines_before_suffix("")
            line_to_parse = line_to_parse[:-1]
            start_index, extracted_whitespace = ParserHelper.extract_spaces_verified(
                line_to_parse, 0
            )
            POGGER.debug(">>line_to_parse>>$<<", line_to_parse)
            (
                did_complete_lrd,
                end_lrd_index,
                parsed_lrd_tuple,
            ) = LinkReferenceDefinitionParseHelper.parse_link_reference_definition(
                parser_state,
                line_to_parse,
                start_index,
                extracted_whitespace,
                is_blank_line,
                "",
                False,
            )
            POGGER.debug(
                ">>parse_link_reference_definition>>was_started>>did_complete_lrd>>$"
                + ">>end_lrd_index>>$>>len(line_to_parse)>>$",
                did_complete_lrd,
                end_lrd_index,
                len(line_to_parse),
            )
            if did_complete_lrd:
                break
        assert is_blank_line is not None, "while loop must have executed at least once."
        assert line_to_parse is not None, "while loop must have executed at least once."
        assert (
            did_complete_lrd is not None
        ), "while loop must have executed at least once."
        assert end_lrd_index is not None, "while loop must have executed at least once."
        return (
            is_blank_line,
            line_to_parse,
            did_complete_lrd,
            end_lrd_index,
            parsed_lrd_tuple,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def handle_link_reference_definition_leaf_block(
        parser_state: ParserState,
        outer_processed: bool,
        position_marker: PositionMarker,
        leaf_token_whitespace: str,
        remaining_line_to_parse: str,
        ignore_link_definition_start: bool,
        pre_tokens: List[MarkdownToken],
        original_line: str,
    ) -> Tuple[bool, Optional[RequeueLineInfo]]:
        """
        Take care of the processing for link reference definitions.
        """
        POGGER.debug(
            "handle_link_reference_definition>>pre_tokens>>$<<",
            pre_tokens,
        )

        if (
            not outer_processed
            and not ignore_link_definition_start
            and not parser_state.token_stack[-1].was_table_block_started
        ):
            POGGER.debug(
                "plflb-process_link_reference_definition>>outer_processed>>$",
                position_marker.text_to_parse[position_marker.index_number :],
            )
            assert (
                parser_state.original_line_to_parse is not None
            ), "Original line must be defined by now."
            (
                outer_processed,
                _,  # did_complete_lrd,
                _,  # did_pause_lrd,
                requeue_line_info,
                new_tokens,
            ) = LinkReferenceDefinitionHelper.process_link_reference_definition(
                parser_state,
                position_marker,
                remaining_line_to_parse,
                leaf_token_whitespace,
                parser_state.original_line_to_parse,
                parser_state.original_stack_depth,
                parser_state.original_document_depth,
                original_line,
            )
            if requeue_line_info:
                outer_processed = True
                POGGER.debug(
                    "plflb-process_link_reference_definition>>outer_processed>>$<lines_to_requeue<$<$",
                    outer_processed,
                    requeue_line_info.lines_to_requeue,
                    len(requeue_line_info.lines_to_requeue),
                )
            else:
                POGGER.debug(
                    "plflb-process_link_reference_definition>>outer_processed>>$<lines_to_requeue<(None)",
                    outer_processed,
                )
        else:
            requeue_line_info, new_tokens = None, []

        POGGER.debug("handle_link_reference_definition>>pre_tokens>>$<<", pre_tokens)
        pre_tokens.extend(new_tokens)
        POGGER.debug("handle_link_reference_definition>>pre_tokens>>$<<", pre_tokens)
        return outer_processed, requeue_line_info

    # pylint: enable=too-many-arguments
